"""Agno Infra Cli

This is the entrypoint for the `agno infra` application.
"""

from pathlib import Path
from typing import List, Optional, cast

import typer

from agno.cli.console import (
    log_active_infra_not_available,
    log_config_not_available_msg,
    print_available_infra,
    print_info,
)
from agno.infra.config import InfraConfig
from agno.utilities.logging import logger, set_log_level_to_debug

infra_cli = typer.Typer(
    name="infra",
    short_help="Manage Agent Infrastructure",
    help="""\b
Use `ag infra [COMMAND]` to create, setup, start or stop your infrastructure.
Run `ag infra [COMMAND] --help` for more info.
""",
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="",
    subcommand_metavar="[COMMAND] [OPTIONS]",
)


@infra_cli.command(short_help="Create a new AgentOS codebase in the current directory.")
def create(
    name: Optional[str] = typer.Option(
        None,
        "-n",
        "--name",
        help="Name of the new AgentOS codebase directory (e.g. `my-agentos-project`).",
        show_default=False,
    ),
    template: Optional[str] = typer.Option(
        None,
        "-t",
        "--template",
        help="Starter template for the Agno AgentOS codebase.",
        show_default=False,
    ),
    url: Optional[str] = typer.Option(
        None,
        "-u",
        "--url",
        help="URL of the starter template.",
        show_default=False,
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Create a new AgentOS Infrastructure project in the current directory using a starter template
    \b
    Examples:
    > ag infra create -t agentos-aws-template                        -> Create an `agentos-aws-template` in the current directory
    > ag infra create -t agentos-aws-template -n my-agentos-template     -> Create an `agentos-aws-template` named `my-agentos-template` in the current directory
    """
    if print_debug_log:
        set_log_level_to_debug()

    from agno.infra.operator import create_infra_from_template

    create_infra_from_template(name=name, template=template, url=url)


@infra_cli.command(short_help="Start the resources for the active AgentOS codebase")
def up(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:INFRA:GROUP:NAME:TYPE",
    ),
    env_filter: Optional[str] = typer.Option(None, "-e", "--env", metavar="", help="Filter the environment to deploy."),
    infra_filter: Optional[str] = typer.Option(None, "-i", "--infra", metavar="", help="Filter the infra to deploy."),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter resources using group name."
    ),
    name_filter: Optional[str] = typer.Option(None, "-n", "--name", metavar="", help="Filter resource using name."),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter resource using type",
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print resources and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip confirmation before deploying resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: Optional[bool] = typer.Option(
        None,
        "-f",
        "--force",
        help="Force create resources where applicable.",
    ),
    pull: Optional[bool] = typer.Option(
        None,
        "-p",
        "--pull",
        help="Pull images where applicable.",
    ),
):
    """\b
    Create resources for the active AgentOS codebase
    Options can be used to limit the resources to create.
      --env     : Env (dev, stg, prd)
      --infra   : Infra type (docker, aws)
      --group   : Group name
      --name    : Resource name
      --type    : Resource type
    \b
    Options can also be provided as a RESOURCE_FILTER in the format: ENV:INFRA:GROUP:NAME:TYPE
    \b
    Examples:
    > `ag infra up`            -> Deploy all resources
    > `ag infra up dev`        -> Deploy all dev resources
    > `ag infra up prd`        -> Deploy all prd resources
    > `ag infra up prd:aws`    -> Deploy all prd aws resources
    > `ag infra up prd:::s3`   -> Deploy prd resources matching name s3
    """
    if print_debug_log:
        set_log_level_to_debug()

    from agno.cli.config import AgnoCliConfig
    from agno.cli.operator import initialize_agno_cli
    from agno.cli.utils import find_compose_files, run_docker_compose_up
    from agno.infra.helpers import get_infra_dir_path
    from agno.infra.operator import setup_infra_config_from_dir, start_infra
    from agno.utilities.resource_filter import parse_resource_filter

    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if not agno_config:
        agno_config = initialize_agno_cli()
        if not agno_config:
            log_config_not_available_msg()
            return
    agno_config = cast(AgnoCliConfig, agno_config)

    # Workspace to start
    infra_to_start: Optional[InfraConfig] = None

    # If there is an existing infra project at current path, use that infra project
    current_path: Path = Path(".").resolve()
    infra_at_current_path: Optional[InfraConfig] = agno_config.get_infra_config_by_path(current_path)
    if infra_at_current_path is not None:
        logger.debug(f"Found Agno Infra project at: {infra_at_current_path.infra_root_path}")
        if str(infra_at_current_path.infra_root_path) != agno_config.active_infra_dir:
            logger.debug(f"Updating active Agno Infra project to {infra_at_current_path.infra_root_path}")
            agno_config.set_active_infra_dir(infra_at_current_path.infra_root_path)
        infra_to_start = infra_at_current_path

    # If there's no existing Infra at current path, check if there's a `infra` dir in the current path
    # In that case setup the Infra
    if infra_to_start is None:
        infra_infra_dir_path = get_infra_dir_path(current_path)
        if infra_infra_dir_path is not None:
            logger.debug(f"Found Infra directory: {infra_infra_dir_path}")
            logger.debug(f"Setting up a Infra at: {current_path}")
            infra_to_start = setup_infra_config_from_dir(infra_root_path=current_path)
            print_info("")

    # If there's no Infra at current path, check if an active Infra exists
    if infra_to_start is None:
        active_infra_config: Optional[InfraConfig] = agno_config.get_active_infra_config()
        # If there's an active Infra, use that Infra
        if active_infra_config is not None:
            infra_to_start = active_infra_config

    # If there's no Infra to start, raise an error showing available Infra
    if infra_to_start is None:
        log_active_infra_not_available()
        avl_infra = agno_config.available_infra
        if avl_infra:
            print_available_infra(avl_infra)

    # Check for docker compose files
    current_dir = Path.cwd()
    compose_files = find_compose_files(current_dir)
    if not compose_files and agno_config.active_infra_dir:
        compose_files = find_compose_files(Path(agno_config.active_infra_dir))
    if infra_to_start is None and not compose_files:
        return
    elif compose_files:
        from agno.cli.console import confirm_yes_no

        logger.info(f"Found Docker Compose files: {[f.name for f in compose_files]}")

        for compose_file in compose_files:
            if not dry_run:
                if not auto_confirm:
                    confirm = confirm_yes_no(f"Run docker compose up for {compose_file.name}?")
                    if not confirm:
                        continue

                # TODO: Use the -f flag logic to force build
                run_docker_compose_up(compose_file, build=True, detached=True)
            else:
                print(f"Would run: docker compose -f {compose_file} up -d --build")
        return

    target_env: Optional[str] = None
    target_infra: Optional[str] = None
    target_group: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    # derive env:infra:name:type:group from ws_filter
    if resource_filter is not None:
        if not isinstance(resource_filter, str):
            raise TypeError(f"Invalid resource_filter. Expected: str, Received: {type(resource_filter)}")
        (
            target_env,
            target_infra,
            target_group,
            target_name,
            target_type,
        ) = parse_resource_filter(resource_filter)

    # derive env:infra:name:type:group from command options
    if target_env is None and env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter
    if target_infra is None and infra_filter is not None and isinstance(infra_filter, str):
        target_infra = infra_filter
    if target_group is None and group_filter is not None and isinstance(group_filter, str):
        target_group = group_filter
    if target_name is None and name_filter is not None and isinstance(name_filter, str):
        target_name = name_filter
    if target_type is None and type_filter is not None and isinstance(type_filter, str):
        target_type = type_filter

    # derive env:infra:name:type:group from defaults
    if target_env is None and infra_to_start and infra_to_start.infra_settings:
        target_env = infra_to_start.infra_settings.default_env if infra_to_start.infra_settings else None
    if target_infra is None and infra_to_start and infra_to_start.infra_settings:
        target_infra = infra_to_start.infra_settings.default_infra if infra_to_start.infra_settings else None

    start_infra(
        infra_config=infra_to_start,  # type: ignore
        target_env=target_env,
        target_infra=target_infra,
        target_group=target_group,
        target_name=target_name,
        target_type=target_type,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
        force=force,
        pull=pull,
    )


@infra_cli.command(short_help="Delete resources for active AgentOS codebase")
def down(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:INFRA:GROUP:NAME:TYPE",
    ),
    env_filter: str = typer.Option(None, "-e", "--env", metavar="", help="Filter the environment to shut down."),
    infra_filter: Optional[str] = typer.Option(
        None, "-i", "--infra", metavar="", help="Filter the infra to shut down."
    ),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter resources using group name."
    ),
    name_filter: Optional[str] = typer.Option(None, "-n", "--name", metavar="", help="Filter resource using name."),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter resource using type",
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print resources and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip the confirmation before deleting resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: bool = typer.Option(
        None,
        "-f",
        "--force",
        help="Force",
    ),
):
    """\b
    Delete resources for the active AgentOS codebase.
    Options can be used to limit the resources to delete.
      --env     : Env (dev, stg, prd)
      --infra   : Infra type (docker, aws)
      --group   : Group name
      --name    : Resource name
      --type    : Resource type
    \b
    Options can also be provided as a RESOURCE_FILTER in the format: ENV:INFRA:GROUP:NAME:TYPE
    \b
    Examples:
    > `ag infra down`            -> Delete all resources
    """
    if print_debug_log:
        set_log_level_to_debug()

    from agno.cli.config import AgnoCliConfig
    from agno.cli.operator import initialize_agno_cli
    from agno.cli.utils import find_compose_files, run_docker_compose_down
    from agno.infra.helpers import get_infra_dir_path
    from agno.infra.operator import setup_infra_config_from_dir, stop_infra
    from agno.utilities.resource_filter import parse_resource_filter

    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if not agno_config:
        agno_config = initialize_agno_cli()
        if not agno_config:
            log_config_not_available_msg()
            return

    # Infra to stop
    infra_to_stop: Optional[InfraConfig] = None

    # If there is an existing Infra at current path, use that Infra
    current_path: Path = Path(".").resolve()
    infra_at_current_path: Optional[InfraConfig] = agno_config.get_infra_config_by_path(current_path)
    if infra_at_current_path is not None:
        logger.debug(f"Found Infra at: {infra_at_current_path.infra_root_path}")
        if str(infra_at_current_path.infra_root_path) != agno_config.active_infra_dir:
            logger.debug(f"Updating active Infra to {infra_at_current_path.infra_root_path}")
            agno_config.set_active_infra_dir(infra_at_current_path.infra_root_path)
        infra_to_stop = infra_at_current_path

    # If there's no existing Infra at current path, check if there's a `infra` dir in the current path
    # In that case setup the Infra
    if infra_to_stop is None:
        infra_infra_dir_path = get_infra_dir_path(current_path)
        if infra_infra_dir_path is not None:
            logger.debug(f"Found infra directory: {infra_infra_dir_path}")
            logger.debug(f"Setting up a Infra at: {current_path}")
            infra_to_stop = setup_infra_config_from_dir(infra_root_path=current_path)
            print_info("")

    # If there's no Infra at current path, check if an active Infra exists
    if infra_to_stop is None:
        active_infra_config: Optional[InfraConfig] = agno_config.get_active_infra_config()
        # If there's an active Infra, use that Infra
        if active_infra_config is not None:
            infra_to_stop = active_infra_config

    # If there's no Infra to stop, raise an error showing available Infra
    if infra_to_stop is None:
        log_active_infra_not_available()
        avl_infra = agno_config.available_infra
        if avl_infra:
            print_available_infra(avl_infra)

    # Check for docker compose files
    current_dir = Path.cwd()
    compose_files = find_compose_files(current_dir)
    if not compose_files and agno_config.active_infra_dir:
        compose_files = find_compose_files(Path(agno_config.active_infra_dir))

    if infra_to_stop is None and not compose_files:
        return
    elif compose_files:
        from agno.cli.console import confirm_yes_no

        logger.info(f"Found Docker Compose files: {[f.name for f in compose_files]}")

        for compose_file in compose_files:
            if not dry_run:
                if not auto_confirm:
                    confirm = confirm_yes_no(f"Run docker compose down for {compose_file.name}?")
                    if not confirm:
                        continue

                run_docker_compose_down(compose_file, remove_volumes=False)
            else:
                print(f"Would run: docker compose -f {compose_file} down")
        return

    target_env: Optional[str] = None
    target_infra: Optional[str] = None
    target_group: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    # derive env:infra:name:type:group from ws_filter
    if resource_filter is not None:
        if not isinstance(resource_filter, str):
            raise TypeError(f"Invalid resource_filter. Expected: str, Received: {type(resource_filter)}")
        (
            target_env,
            target_infra,
            target_group,
            target_name,
            target_type,
        ) = parse_resource_filter(resource_filter)

    # derive env:infra:name:type:group from command options
    if target_env is None and env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter
    if target_infra is None and infra_filter is not None and isinstance(infra_filter, str):
        target_infra = infra_filter
    if target_group is None and group_filter is not None and isinstance(group_filter, str):
        target_group = group_filter
    if target_name is None and name_filter is not None and isinstance(name_filter, str):
        target_name = name_filter
    if target_type is None and type_filter is not None and isinstance(type_filter, str):
        target_type = type_filter

    # derive env:infra:name:type:group from defaults
    if target_env is None and infra_to_stop and infra_to_stop.infra_settings:
        target_env = infra_to_stop.infra_settings.default_env if infra_to_stop.infra_settings else None
    if target_infra is None and infra_to_stop and infra_to_stop.infra_settings:
        target_infra = infra_to_stop.infra_settings.default_infra if infra_to_stop.infra_settings else None

    stop_infra(
        infra_config=infra_to_stop,  # type: ignore
        target_env=target_env,
        target_infra=target_infra,
        target_group=target_group,
        target_name=target_name,
        target_type=target_type,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
        force=force,
    )


@infra_cli.command(short_help="Update resources for active AgentOS codebase")
def patch(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:INFRA:GROUP:NAME:TYPE",
    ),
    env_filter: str = typer.Option(None, "-e", "--env", metavar="", help="Filter the environment to patch."),
    infra_filter: Optional[str] = typer.Option(None, "-i", "--infra", metavar="", help="Filter the infra to patch."),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter resources using group name."
    ),
    name_filter: Optional[str] = typer.Option(None, "-n", "--name", metavar="", help="Filter resource using name."),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter resource using type",
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print resources and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip the confirmation before patching resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: bool = typer.Option(
        None,
        "-f",
        "--force",
        help="Force",
    ),
    pull: Optional[bool] = typer.Option(
        None,
        "-p",
        "--pull",
        help="Pull images where applicable.",
    ),
):
    """\b
    Update resources for the active AgentOS codebase.
    Options can be used to limit the resources to update.
      --env     : Env (dev, stg, prd)
      --infra   : Infra type (docker, aws)
      --group   : Group name
      --name    : Resource name
      --type    : Resource type
    \b
    Options can also be provided as a RESOURCE_FILTER in the format: ENV:INFRA:GROUP:NAME:TYPE
    Examples:
    \b
    > `ag infra patch`           -> Patch all resources
    """
    if print_debug_log:
        set_log_level_to_debug()

    from agno.cli.config import AgnoCliConfig
    from agno.cli.operator import initialize_agno_cli
    from agno.infra.helpers import get_infra_dir_path
    from agno.infra.operator import setup_infra_config_from_dir, update_infra
    from agno.utilities.resource_filter import parse_resource_filter

    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if not agno_config:
        agno_config = initialize_agno_cli()
        if not agno_config:
            log_config_not_available_msg()
            return

    # Infra to patch
    infra_to_patch: Optional[InfraConfig] = None

    # If there is an existing Infra at current path, use that Infra
    current_path: Path = Path(".").resolve()
    infra_at_current_path: Optional[InfraConfig] = agno_config.get_infra_config_by_path(current_path)
    if infra_at_current_path is not None:
        logger.debug(f"Found infra at: {infra_at_current_path.infra_root_path}")
        if str(infra_at_current_path.infra_root_path) != agno_config.active_infra_dir:
            logger.debug(f"Updating active infra to {infra_at_current_path.infra_root_path}")
            agno_config.set_active_infra_dir(infra_at_current_path.infra_root_path)
        infra_to_patch = infra_at_current_path

    # If there's no existing Infra at current path, check if there's a `infra` dir in the current path
    # In that case setup the Infra
    if infra_to_patch is None:
        infra_infra_dir_path = get_infra_dir_path(current_path)
        if infra_infra_dir_path is not None:
            logger.debug(f"Found infra directory: {infra_infra_dir_path}")
            logger.debug(f"Setting up a Infra at: {current_path}")
            infra_to_patch = setup_infra_config_from_dir(infra_root_path=current_path)
            print_info("")

    # If there's no Infra at current path, check if an active Infra exists
    if infra_to_patch is None:
        active_infra_config: Optional[InfraConfig] = agno_config.get_active_infra_config()
        # If there's an active Infra, use that Infra
        if active_infra_config is not None:
            infra_to_patch = active_infra_config

    # If there's no Infra to patch, raise an error showing available Infra
    if infra_to_patch is None:
        log_active_infra_not_available()
        avl_infra = agno_config.available_infra
        if avl_infra:
            print_available_infra(avl_infra)
        return

    target_env: Optional[str] = None
    target_infra: Optional[str] = None
    target_group: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    # derive env:infra:name:type:group from ws_filter
    if resource_filter is not None:
        if not isinstance(resource_filter, str):
            raise TypeError(f"Invalid resource_filter. Expected: str, Received: {type(resource_filter)}")
        (
            target_env,
            target_infra,
            target_group,
            target_name,
            target_type,
        ) = parse_resource_filter(resource_filter)

    # derive env:infra:name:type:group from command options
    if target_env is None and env_filter is not None and isinstance(env_filter, str):
        target_env = env_filter
    if target_infra is None and infra_filter is not None and isinstance(infra_filter, str):
        target_infra = infra_filter
    if target_group is None and group_filter is not None and isinstance(group_filter, str):
        target_group = group_filter
    if target_name is None and name_filter is not None and isinstance(name_filter, str):
        target_name = name_filter
    if target_type is None and type_filter is not None and isinstance(type_filter, str):
        target_type = type_filter

    # derive env:infra:name:type:group from defaults
    if target_env is None:
        target_env = infra_to_patch.infra_settings.default_env if infra_to_patch.infra_settings else None
    if target_infra is None:
        target_infra = infra_to_patch.infra_settings.default_infra if infra_to_patch.infra_settings else None

    update_infra(
        infra_config=infra_to_patch,
        target_env=target_env,
        target_infra=target_infra,
        target_group=target_group,
        target_name=target_name,
        target_type=target_type,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
        force=force,
        pull=pull,
    )


@infra_cli.command(short_help="Restart resources for active AgentOS codebase")
def restart(
    resource_filter: Optional[str] = typer.Argument(
        None,
        help="Resource filter. Format - ENV:INFRA:GROUP:NAME:TYPE",
    ),
    env_filter: str = typer.Option(None, "-e", "--env", metavar="", help="Filter the environment to restart."),
    infra_filter: Optional[str] = typer.Option(None, "-i", "--infra", metavar="", help="Filter the infra to restart."),
    group_filter: Optional[str] = typer.Option(
        None, "-g", "--group", metavar="", help="Filter resources using group name."
    ),
    name_filter: Optional[str] = typer.Option(None, "-n", "--name", metavar="", help="Filter resource using name."),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter resource using type",
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print resources and exit.",
    ),
    auto_confirm: bool = typer.Option(
        False,
        "-y",
        "--yes",
        help="Skip the confirmation before restarting resources.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
    force: bool = typer.Option(
        None,
        "-f",
        "--force",
        help="Force",
    ),
    pull: Optional[bool] = typer.Option(
        None,
        "-p",
        "--pull",
        help="Pull images where applicable.",
    ),
):
    """\b
    Restarts the active AgentOS codebase. i.e. runs `ag infra down` and then `ag infra up`.

    \b
    Examples:
    > `ag infra restart`
    """
    if print_debug_log:
        set_log_level_to_debug()

    from time import sleep

    down(
        resource_filter=resource_filter,
        env_filter=env_filter,
        group_filter=group_filter,
        infra_filter=infra_filter,
        name_filter=name_filter,
        type_filter=type_filter,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
        print_debug_log=print_debug_log,
        force=force,
    )
    print_info("Sleeping for 2 seconds..")
    sleep(2)
    up(
        resource_filter=resource_filter,
        env_filter=env_filter,
        infra_filter=infra_filter,
        group_filter=group_filter,
        name_filter=name_filter,
        type_filter=type_filter,
        dry_run=dry_run,
        auto_confirm=auto_confirm,
        print_debug_log=print_debug_log,
        force=force,
        pull=pull,
    )


@infra_cli.command(short_help="Prints active AgentOS codebase config")
def config(
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Prints the active AgentOS codebase config

    \b
    Examples:
    $ `ag infra config`         -> Print the active Infra config
    """
    if print_debug_log:
        set_log_level_to_debug()

    from agno.cli.config import AgnoCliConfig
    from agno.cli.operator import initialize_agno_cli
    from agno.utilities.load_env import load_env

    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if not agno_config:
        agno_config = initialize_agno_cli()
        if not agno_config:
            log_config_not_available_msg()
            return

    active_infra_config: Optional[InfraConfig] = agno_config.get_active_infra_config()
    if active_infra_config is None:
        log_active_infra_not_available()
        avl_infra = agno_config.available_infra
        if avl_infra:
            print_available_infra(avl_infra)
        return

    # Load environment from .env
    load_env(
        dotenv_dir=active_infra_config.infra_root_path,
    )
    print_info(active_infra_config.model_dump_json(include={"infra_name", "infra_root_path"}, indent=2))


@infra_cli.command(short_help="Delete AgentOS codebase infra record")
def delete(
    infra_name: Optional[str] = typer.Option(None, "-infra", help="Name of the AgentOS codebase to delete"),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Deletes the AgentOS codebase record from agno_infra.
    NOTE: Does not delete any physical files.

    \b
    Examples:
    $ `ag infra delete`         -> Delete the active AgentOS codebase from Agno
    """
    if print_debug_log:
        set_log_level_to_debug()

    from agno.cli.config import AgnoCliConfig
    from agno.cli.operator import initialize_agno_cli
    from agno.infra.operator import delete_infra

    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if not agno_config:
        agno_config = initialize_agno_cli()
        if not agno_config:
            log_config_not_available_msg()
            return

    infra_to_delete: List[Path] = []
    # Delete Infra by name if provided
    if infra_name is not None:
        infra_config = agno_config.get_infra_config_by_dir_name(infra_name)
        if infra_config is None:
            logger.error(f"AgentOS codebase {infra_name} not found")
            return
        infra_to_delete.append(infra_config.infra_root_path)
    else:
        # By default, we assume this command is run for the active Infra
        if agno_config.active_infra_dir is not None:
            infra_to_delete.append(Path(agno_config.active_infra_dir))

    delete_infra(agno_config, infra_to_delete)



from typing import Any, Optional

from agno.aws.api_client import AwsApiClient
from agno.base.resource import InfraResource
from agno.cli.console import print_info
from agno.utilities.logging import logger


class AwsResource(InfraResource):
    """Base class for AWS Resources."""

    service_name: str
    service_client: Optional[Any] = None
    service_resource: Optional[Any] = None

    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None

    aws_client: Optional[AwsApiClient] = None

    def get_aws_region(self) -> Optional[str]:
        # Priority 1: Use aws_region from resource
        if self.aws_region:
            return self.aws_region

        # Priority 2: Get aws_region from os settings
        if self.infra_settings is not None and self.infra_settings.aws_region is not None:
            self.aws_region = self.infra_settings.aws_region
            return self.aws_region

        # Priority 3: Get aws_region from env
        from os import getenv

        from agno.constants import AWS_REGION_ENV_VAR

        aws_region_env = getenv(AWS_REGION_ENV_VAR)
        if aws_region_env is not None:
            logger.debug(f"{AWS_REGION_ENV_VAR}: {aws_region_env}")
            self.aws_region = aws_region_env
        return self.aws_region

    def get_aws_profile(self) -> Optional[str]:
        # Priority 1: Use aws_region from resource
        if self.aws_profile:
            return self.aws_profile

        # Priority 2: Get aws_profile from os settings
        if self.infra_settings is not None and self.infra_settings.aws_profile is not None:
            self.aws_profile = self.infra_settings.aws_profile
            return self.aws_profile

        # Priority 3: Get aws_profile from env
        from os import getenv

        from agno.constants import AWS_PROFILE_ENV_VAR

        aws_profile_env = getenv(AWS_PROFILE_ENV_VAR)
        if aws_profile_env is not None:
            logger.debug(f"{AWS_PROFILE_ENV_VAR}: {aws_profile_env}")
            self.aws_profile = aws_profile_env
        return self.aws_profile

    def get_service_client(self, aws_client: AwsApiClient):
        from boto3 import session

        if self.service_client is None:
            boto3_session: session = aws_client.boto3_session
            self.service_client = boto3_session.client(service_name=self.service_name)
        return self.service_client

    def get_service_resource(self, aws_client: AwsApiClient):
        from boto3 import session

        if self.service_resource is None:
            boto3_session: session = aws_client.boto3_session
            self.service_resource = boto3_session.resource(service_name=self.service_name)
        return self.service_resource

    def get_aws_client(self) -> AwsApiClient:
        if self.aws_client is not None:
            return self.aws_client
        self.aws_client = AwsApiClient(aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile())
        return self.aws_client

    def _read(self, aws_client: AwsApiClient) -> Any:
        logger.warning(f"@_read method not defined for {self.get_resource_name()}")
        return True

    def read(self, aws_client: Optional[AwsApiClient] = None) -> Any:
        """Reads the resource from Aws"""
        # Step 1: Use cached value if available
        if self.use_cache and self.active_resource is not None:
            return self.active_resource

        # Step 2: Skip resource creation if skip_read = True
        if self.skip_read:
            print_info(f"Skipping read: {self.get_resource_name()}")
            return True

        # Step 3: Read resource
        client: AwsApiClient = aws_client or self.get_aws_client()
        return self._read(client)

    def is_active(self, aws_client: AwsApiClient) -> bool:
        """Returns True if the resource is active on Aws"""
        _resource = self.read(aws_client=aws_client)
        return True if _resource is not None else False

    def _create(self, aws_client: AwsApiClient) -> bool:
        logger.warning(f"@_create method not defined for {self.get_resource_name()}")
        return True

    def create(self, aws_client: Optional[AwsApiClient] = None) -> bool:
        """Creates the resource on Aws"""

        # Step 1: Skip resource creation if skip_create = True
        if self.skip_create:
            print_info(f"Skipping create: {self.get_resource_name()}")
            return True

        # Step 2: Check if resource is active and use_cache = True
        client: AwsApiClient = aws_client or self.get_aws_client()
        if self.use_cache and self.is_active(client):
            self.resource_created = True
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} already exists")
        # Step 3: Create the resource
        else:
            self.resource_created = self._create(client)
            if self.resource_created:
                print_info(f"{self.get_resource_type()}: {self.get_resource_name()} created")

        # Step 4: Run post create steps
        if self.resource_created:
            if self.save_output:
                self.save_output_file()
            logger.debug(f"Running post-create for {self.get_resource_type()}: {self.get_resource_name()}")
            return self.post_create(client)
        logger.error(f"Failed to create {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_created

    def post_create(self, aws_client: AwsApiClient) -> bool:
        return True

    def _update(self, aws_client: AwsApiClient) -> Any:
        logger.warning(f"@_update method not defined for {self.get_resource_name()}")
        return True

    def update(self, aws_client: Optional[AwsApiClient] = None) -> bool:
        """Updates the resource on Aws"""

        # Step 1: Skip resource update if skip_update = True
        if self.skip_update:
            print_info(f"Skipping update: {self.get_resource_name()}")
            return True

        # Step 2: Update the resource
        client: AwsApiClient = aws_client or self.get_aws_client()
        if self.is_active(client):
            self.resource_updated = self._update(client)
        else:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} does not exist")
            return True

        # Step 3: Run post update steps
        if self.resource_updated:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} updated")
            if self.save_output:
                self.save_output_file()
            logger.debug(f"Running post-update for {self.get_resource_type()}: {self.get_resource_name()}")
            return self.post_update(client)
        logger.error(f"Failed to update {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_updated

    def post_update(self, aws_client: AwsApiClient) -> bool:
        return True

    def _delete(self, aws_client: AwsApiClient) -> Any:
        logger.warning(f"@_delete method not defined for {self.get_resource_name()}")
        return True

    def delete(self, aws_client: Optional[AwsApiClient] = None) -> bool:
        """Deletes the resource from Aws"""

        # Step 1: Skip resource deletion if skip_delete = True
        if self.skip_delete:
            print_info(f"Skipping delete: {self.get_resource_name()}")
            return True

        # Step 2: Delete the resource
        client: AwsApiClient = aws_client or self.get_aws_client()
        if self.is_active(client):
            self.resource_deleted = self._delete(client)
        else:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} does not exist")
            return True

        # Step 3: Run post delete steps
        if self.resource_deleted:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} deleted")
            if self.save_output:
                self.delete_output_file()
            logger.debug(f"Running post-delete for {self.get_resource_type()}: {self.get_resource_name()}.")
            return self.post_delete(client)
        logger.error(f"Failed to delete {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_deleted

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        return True



from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from agno.base.app import InfraApp
from agno.base.context import ContainerContext
from agno.docker.context import DockerBuildContext
from agno.utilities.logging import logger

if TYPE_CHECKING:
    from agno.docker.resource.base import DockerResource


class DockerApp(InfraApp):
    # -*- OS Configuration
    # Path to the infra directory inside the container
    infra_dir_container_path: str = "/app"
    # Mount the infra directory from host machine to the container
    mount_infra_dir: bool = False

    # -*- App Volume
    # Create a volume for container storage
    create_volume: bool = False
    # If volume_dir is provided, mount this directory RELATIVE to the workspace_root
    # from the host machine to the volume_container_path
    volume_dir: Optional[str] = None
    # Otherwise, mount a volume named volume_name to the container
    # If volume_name is not provided, use {app-name}-volume
    volume_name: Optional[str] = None
    # Path to mount the volume inside the container
    volume_container_path: str = "/mnt/app"

    # -*- Resources Volume
    # Mount a read-only directory from host machine to the container
    mount_resources: bool = False
    # Resources directory relative to the workspace_root
    resources_dir: str = "workspace/resources"
    # Path to mount the resources_dir
    resources_dir_container_path: str = "/mnt/resources"

    # -*- Agno Volume
    # Mount ~/.config/ag directory from host machine to the container
    mount_agno_config: bool = True

    # -*- Container Configuration
    container_name: Optional[str] = None
    container_labels: Optional[Dict[str, str]] = None
    # Run container in the background and return a Container object
    container_detach: bool = True
    # Enable auto-removal of the container on daemon side when the container’s process exits
    container_auto_remove: bool = True
    # Remove the container when it has finished running. Default: True
    container_remove: bool = True
    # Username or UID to run commands as inside the container
    container_user: Optional[Union[str, int]] = None
    # Keep STDIN open even if not attached
    container_stdin_open: bool = True
    # Return logs from STDOUT when container_detach=False
    container_stdout: Optional[bool] = True
    # Return logs from STDERR when container_detach=False
    container_stderr: Optional[bool] = True
    container_tty: bool = True
    # Specify a test to perform to check that the container is healthy
    container_healthcheck: Optional[Dict[str, Any]] = None
    # Optional hostname for the container
    container_hostname: Optional[str] = None
    # Platform in the format os[/arch[/variant]]
    container_platform: Optional[str] = None
    # Path to the working directory
    container_working_dir: Optional[str] = None
    # Restart the container when it exits. Configured as a dictionary with keys:
    # Name: One of on-failure, or always.
    # MaximumRetryCount: Number of times to restart the container on failure.
    # For example: {"Name": "on-failure", "MaximumRetryCount": 5}
    container_restart_policy: Optional[Dict[str, Any]] = None
    # Add volumes to DockerContainer
    # container_volumes is a dictionary which adds the volumes to mount
    # inside the container. The key is either the host path or a volume name,
    # and the value is a dictionary with 2 keys:
    #   bind - The path to mount the volume inside the container
    #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
    # For example:
    # {
    #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
    #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
    # }
    container_volumes: Optional[Dict[str, dict]] = None
    # Add ports to DockerContainer
    # The keys of the dictionary are the ports to bind inside the container,
    # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.
    # The values of the dictionary are the corresponding ports to open on the host, which can be either:
    #   - The port number, as an integer.
    #     For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
    #   - None, to assign a random host port. For example, {'2222/tcp': None}.
    #   - A tuple of (address, port) if you want to specify the host interface.
    #     For example, {'1111/tcp': ('127.0.0.1', 1111)}.
    #   - A list of integers, if you want to bind multiple host ports to a single container port.
    #     For example, {'1111/tcp': [1234, 4567]}.
    container_ports: Optional[Dict[str, Any]] = None

    def get_container_name(self) -> str:
        return self.container_name or self.get_app_name()

    def get_container_context(self) -> Optional[ContainerContext]:
        logger.debug("Building ContainerContext")

        if self.container_context is not None:  # type: ignore
            return self.container_context  # type: ignore

        infra_name = self.infra_name
        if infra_name is None:
            raise Exception("Could not determine infra_name")

        infra_root_in_container = self.infra_dir_container_path
        if infra_root_in_container is None:
            raise Exception("Could not determine infra_root in container")

        infra_parent_paths = infra_root_in_container.split("/")[0:-1]
        infra_parent_in_container = "/".join(infra_parent_paths)

        self.container_context = ContainerContext(
            infra_name=infra_name,
            infra_root=infra_root_in_container,
            infra_parent=infra_parent_in_container,
        )

        if self.requirements_file is not None:
            self.container_context.requirements_file = f"{infra_root_in_container}/{self.requirements_file}"  # type: ignore

        return self.container_context

    def get_container_env(self, container_context: ContainerContext) -> Dict[str, str]:
        from agno.constants import (
            AGNO_INFRA_ROOT,
            AGNO_RUNTIME_ENV_VAR,
            PYTHONPATH_ENV_VAR,
            REQUIREMENTS_FILE_PATH_ENV_VAR,
        )

        # Container Environment
        container_env: Dict[str, str] = self.container_env or {}
        container_env.update(
            {
                "INSTALL_REQUIREMENTS": str(self.install_requirements),
                "MOUNT_RESOURCES": str(self.mount_resources),
                "MOUNT_INFRA_DIR": str(self.mount_infra_dir),
                "PRINT_ENV_ON_LOAD": str(self.print_env_on_load),
                "RESOURCES_DIR_CONTAINER_PATH": str(self.resources_dir_container_path),
                AGNO_RUNTIME_ENV_VAR: "docker",
                REQUIREMENTS_FILE_PATH_ENV_VAR: container_context.requirements_file or "",
                AGNO_INFRA_ROOT: container_context.infra_root or "",
            }
        )

        if self.set_python_path:
            python_path = self.python_path
            if python_path is None:
                python_path = container_context.infra_root
                if self.mount_resources and self.resources_dir_container_path is not None:
                    python_path = "{}:{}".format(python_path, self.resources_dir_container_path)
                if self.add_python_paths is not None:
                    python_path = "{}:{}".format(python_path, ":".join(self.add_python_paths))
            if python_path is not None:
                container_env[PYTHONPATH_ENV_VAR] = python_path

        # Update the container env using env_file
        env_data_from_file = self.get_env_file_data()
        if env_data_from_file is not None:
            container_env.update({k: str(v) for k, v in env_data_from_file.items() if v is not None})

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_file_data()
        if secret_data_from_file is not None:
            container_env.update({k: str(v) for k, v in secret_data_from_file.items() if v is not None})

        # Update the container env with user provided env_vars
        # this overwrites any existing variables with the same key
        if self.env_vars is not None and isinstance(self.env_vars, dict):
            container_env.update({k: str(v) for k, v in self.env_vars.items() if v is not None})

        # logger.debug("Container Environment: {}".format(container_env))
        return container_env

    def get_container_volumes(self, container_context: ContainerContext) -> Dict[str, dict]:
        from agno.utilities.defaults import get_default_volume_name

        if self.infra_root is None:
            logger.error("Invalid infra_root")
            return {}

        # container_volumes is a dictionary which configures the volumes to mount
        # inside the container. The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        container_volumes = self.container_volumes or {}

        # Create Workspace Volume
        if self.mount_infra_dir:
            infra_root_in_container = container_context.infra_root
            infra_root_on_host = str(self.infra_root)
            logger.debug(f"Mounting: {infra_root_on_host}")
            logger.debug(f"      to: {infra_root_in_container}")
            container_volumes[infra_root_on_host] = {
                "bind": infra_root_in_container,
                "mode": "rw",
            }

        # Create App Volume
        if self.create_volume:
            volume_host = self.volume_name or get_default_volume_name(self.get_app_name())
            if self.volume_dir is not None:
                volume_host = str(self.infra_root.joinpath(self.volume_dir))
            logger.debug(f"Mounting: {volume_host}")
            logger.debug(f"      to: {self.volume_container_path}")
            container_volumes[volume_host] = {
                "bind": self.volume_container_path,
                "mode": "rw",
            }

        # Create Resources Volume
        if self.mount_resources:
            resources_dir_path = str(self.infra_root.joinpath(self.resources_dir))
            logger.debug(f"Mounting: {resources_dir_path}")
            logger.debug(f"      to: {self.resources_dir_container_path}")
            container_volumes[resources_dir_path] = {
                "bind": self.resources_dir_container_path,
                "mode": "ro",
            }

        # Add ~/.config/ag as a volume
        if self.mount_agno_config:
            agno_config_host_path = str(Path.home().joinpath(".config/ag"))
            agno_config_container_path = f"{self.infra_dir_container_path}/.config/ag"
            logger.debug(f"Mounting: {agno_config_host_path}")
            logger.debug(f"      to: {agno_config_container_path}")
            container_volumes[agno_config_host_path] = {
                "bind": agno_config_container_path,
                "mode": "ro",
            }

        return container_volumes

    def get_container_ports(self) -> Dict[str, int]:
        # container_ports is a dictionary which configures the ports to bind
        # inside the container. The key is the port to bind inside the container
        #   either as an integer or a string in the form port/protocol
        # and the value is the corresponding port to open on the host.
        # For example:
        #   {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
        container_ports: Dict[str, int] = self.container_ports or {}

        if self.open_port:
            _container_port = self.container_port or self.port_number
            _host_port = self.host_port or self.port_number
            container_ports[str(_container_port)] = _host_port

        return container_ports

    def get_container_command(self) -> Optional[List[str]]:
        if isinstance(self.command, str):
            return self.command.strip().split(" ")
        return self.command

    def build_resources(self, build_context: DockerBuildContext) -> List["DockerResource"]:
        from agno.docker.resource.base import DockerResource
        from agno.docker.resource.container import DockerContainer
        from agno.docker.resource.network import DockerNetwork

        logger.debug(f"------------ Building {self.get_app_name()} ------------")
        # -*- Get Container Context
        container_context: Optional[ContainerContext] = self.get_container_context()
        if container_context is None:
            raise Exception("Could not build ContainerContext")
        logger.debug(f"ContainerContext: {container_context.model_dump_json(indent=2)}")

        # -*- Get Container Environment
        container_env: Dict[str, str] = self.get_container_env(container_context=container_context)

        # -*- Get Container Volumes
        container_volumes = self.get_container_volumes(container_context=container_context)

        # -*- Get Container Ports
        container_ports: Dict[str, int] = self.get_container_ports()

        # -*- Get Container Command
        container_cmd: Optional[List[str]] = self.get_container_command()
        if container_cmd:
            logger.debug("Command: {}".format(" ".join(container_cmd)))

        # -*- Build the DockerContainer for this App
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=self.get_image_str(),
            entrypoint=self.entrypoint,
            command=" ".join(container_cmd) if container_cmd is not None else None,
            detach=self.container_detach,
            auto_remove=self.container_auto_remove if not self.debug_mode else False,
            remove=self.container_remove if not self.debug_mode else False,
            healthcheck=self.container_healthcheck,
            hostname=self.container_hostname,
            labels=self.container_labels,
            environment=container_env,
            network=build_context.network,
            platform=self.container_platform,
            ports=container_ports if len(container_ports) > 0 else None,
            restart_policy=self.container_restart_policy,
            stdin_open=self.container_stdin_open,
            stderr=self.container_stderr,
            stdout=self.container_stdout,
            tty=self.container_tty,
            user=self.container_user,
            volumes=container_volumes if len(container_volumes) > 0 else None,
            working_dir=self.container_working_dir,
            use_cache=self.use_cache,
        )

        # -*- List of DockerResources created by this App
        app_resources: List[DockerResource] = []
        if self.image:
            app_resources.append(self.image)
        app_resources.extend(
            [
                DockerNetwork(name=build_context.network),
                docker_container,
            ]
        )

        logger.debug(f"------------ {self.get_app_name()} Built ------------")
        return app_resources

    def get_infra_resources(self) -> Optional[Any]:
        from agno.docker.resources import DockerResources

        return DockerResources(
            name=self.get_app_name(),
            apps=[self],
        )



import logging
from os import getenv
from typing import Any, Optional

from rich.logging import RichHandler
from rich.text import Text

LOGGER_NAME = "agno_infra"

# Define custom styles for different log sources
LOG_STYLES = {
    "warning": "magenta",
    "error": "red",
    "exception": "red",
    "debug": "green",
    "info": "blue",
}


class ColoredRichHandler(RichHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_level_text(self, record: logging.LogRecord) -> Text:
        # Return empty Text if message is empty
        if not record.msg:
            return Text("")

        level_name = record.levelname.lower()
        color = LOG_STYLES[level_name]
        return Text(record.levelname, style=color)


class AgnoLogger(logging.Logger):
    def __init__(self, name: str, level: int = logging.NOTSET):
        super().__init__(name, level)

    def debug(self, msg: str, center: bool = False, symbol: str = "*", *args, **kwargs):
        if center:
            msg = center_header(str(msg), symbol)
        super().debug(msg, *args, **kwargs)

    def info(self, msg: str, center: bool = False, symbol: str = "*", *args, **kwargs):
        if center:
            msg = center_header(str(msg), symbol)
        super().info(msg, *args, **kwargs)


def build_logger(logger_name: str, source_type: Optional[str] = None) -> Any:
    # Set the custom logger class as the default for this logger
    logging.setLoggerClass(AgnoLogger)

    # Create logger with custom class
    _logger = logging.getLogger(logger_name)

    # Reset logger class to default to avoid affecting other loggers
    logging.setLoggerClass(logging.Logger)

    # https://rich.readthedocs.io/en/latest/reference/logging.html#rich.logging.RichHandler
    # https://rich.readthedocs.io/en/latest/logging.html#handle-exceptions
    rich_handler = ColoredRichHandler(
        show_time=False,
        rich_tracebacks=False,
        show_path=True if getenv("AGNO_API_RUNTIME") == "dev" else False,
        tracebacks_show_locals=False,
    )
    rich_handler.setFormatter(
        logging.Formatter(
            fmt="%(message)s",
            datefmt="[%X]",
        )
    )

    _logger.addHandler(rich_handler)
    _logger.setLevel(logging.INFO)
    _logger.propagate = False
    return _logger


logger: AgnoLogger = build_logger(LOGGER_NAME)

debug_on: bool = False


def set_log_level_to_debug():
    _logger = logging.getLogger(LOGGER_NAME)
    _logger.setLevel(logging.DEBUG)

    global debug_on
    debug_on = True


def set_log_level_to_info():
    _logger = logging.getLogger(LOGGER_NAME)
    _logger.setLevel(logging.INFO)

    global debug_on
    debug_on = False


def center_header(message: str, symbol: str = "*") -> str:
    try:
        import shutil

        terminal_width = shutil.get_terminal_size().columns
    except Exception:
        terminal_width = 80  # fallback width

    header = f" {message} "
    return f"{header.center(terminal_width - 20, symbol)}"


def log_debug(msg, center: bool = False, symbol: str = "*", *args, **kwargs):
    global logger
    global debug_on

    if debug_on:
        logger.debug(msg, center, symbol, *args, **kwargs)


def log_info(msg, center: bool = False, symbol: str = "*", *args, **kwargs):
    global logger
    logger.info(msg, center, symbol, *args, **kwargs)


def log_warning(msg, *args, **kwargs):
    global logger
    logger.warning(msg, *args, **kwargs)


def log_error(msg, *args, **kwargs):
    global logger
    logger.error(msg, *args, **kwargs)


def log_exception(msg, *args, **kwargs):
    global logger
    logger.exception(msg, *args, **kwargs)



from typing import Any, Dict, Optional

from agno.aws.resource.base import AwsResource
from agno.utilities.logging import logger


class EcsVolume(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html
    """

    resource_type: Optional[str] = "EcsVolume"
    service_name: str = "ecs"

    name: str
    host: Optional[Dict[str, Any]] = None
    docker_volume_configuration: Optional[Dict[str, Any]] = None
    efs_volume_configuration: Optional[Dict[str, Any]] = None
    fsx_windows_file_server_volume_configuration: Optional[Dict[str, Any]] = None

    def get_volume_definition(self) -> Dict[str, Any]:
        volume_definition: Dict[str, Any] = {}

        if self.name is not None:
            volume_definition["name"] = self.name
        if self.host is not None:
            volume_definition["host"] = self.host
        if self.docker_volume_configuration is not None:
            volume_definition["dockerVolumeConfiguration"] = self.docker_volume_configuration
        if self.efs_volume_configuration is not None:
            volume_definition["efsVolumeConfiguration"] = self.efs_volume_configuration
        if self.fsx_windows_file_server_volume_configuration is not None:
            volume_definition["fsxWindowsFileServerVolumeConfiguration"] = (
                self.fsx_windows_file_server_volume_configuration
            )

        return volume_definition

    def volume_definition_up_to_date(self, volume_definition: Dict[str, Any]) -> bool:
        if self.name is not None:
            if volume_definition.get("name") != self.name:
                logger.debug("{} != {}".format(self.name, volume_definition.get("name")))
                return False
        if self.host is not None:
            if volume_definition.get("host") != self.host:
                logger.debug("{} != {}".format(self.host, volume_definition.get("host")))
                return False
        if self.docker_volume_configuration is not None:
            if volume_definition.get("dockerVolumeConfiguration") != self.docker_volume_configuration:
                logger.debug(
                    "{} != {}".format(
                        self.docker_volume_configuration,
                        volume_definition.get("dockerVolumeConfiguration"),
                    )
                )
                return False
        if self.efs_volume_configuration is not None:
            if volume_definition.get("efsVolumeConfiguration") != self.efs_volume_configuration:
                logger.debug(
                    "{} != {}".format(
                        self.efs_volume_configuration,
                        volume_definition.get("efsVolumeConfiguration"),
                    )
                )
                return False
        if self.fsx_windows_file_server_volume_configuration is not None:
            if (
                volume_definition.get("fsxWindowsFileServerVolumeConfiguration")
                != self.fsx_windows_file_server_volume_configuration
            ):
                logger.debug(
                    "{} != {}".format(
                        self.fsx_windows_file_server_volume_configuration,
                        volume_definition.get("fsxWindowsFileServerVolumeConfiguration"),
                    )
                )
                return False

        return True



from typing import Any, Dict, List, Optional

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.cli.console import print_info
from agno.utilities.logging import logger


class EcsCluster(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html
    """

    resource_type: Optional[str] = "EcsCluster"
    service_name: str = "ecs"

    # Name of the cluster.
    name: str
    # Name for the cluster.
    # Use name if not provided.
    ecs_cluster_name: Optional[str] = None

    tags: Optional[List[Dict[str, str]]] = None
    # The setting to use when creating a cluster.
    settings: Optional[List[Dict[str, Any]]] = None
    # The execute command configuration for the cluster.
    configuration: Optional[Dict[str, Any]] = None
    # The short name of one or more capacity providers to associate with the cluster.
    # A capacity provider must be associated with a cluster before it can be included as part of the default capacity
    # provider strategy of the cluster or used in a capacity provider strategy when calling the CreateService/RunTask.
    capacity_providers: Optional[List[str]] = None
    # The capacity provider strategy to set as the default for the cluster. After a default capacity provider strategy
    # is set for a cluster, when you call the RunTask or CreateService APIs with no capacity provider strategy or
    # launch type specified, the default capacity provider strategy for the cluster is used.
    default_capacity_provider_strategy: Optional[List[Dict[str, Any]]] = None
    # Use this parameter to set a default Service Connect namespace.
    # After you set a default Service Connect namespace, any new services with Service Connect turned on that are
    # created in the cluster are added as client services in the namespace.
    service_connect_namespace: Optional[str] = None

    def get_ecs_cluster_name(self):
        return self.ecs_cluster_name or self.name

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the EcsCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.tags is not None:
            not_null_args["tags"] = self.tags
        if self.settings is not None:
            not_null_args["settings"] = self.settings
        if self.configuration is not None:
            not_null_args["configuration"] = self.configuration
        if self.capacity_providers is not None:
            not_null_args["capacityProviders"] = self.capacity_providers
        if self.default_capacity_provider_strategy is not None:
            not_null_args["defaultCapacityProviderStrategy"] = self.default_capacity_provider_strategy
        if self.service_connect_namespace is not None:
            not_null_args["serviceConnectDefaults"] = {
                "namespace": self.service_connect_namespace,
            }

        # Create EcsCluster
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_cluster(
                clusterName=self.get_ecs_cluster_name(),
                **not_null_args,
            )
            logger.debug(f"EcsCluster: {create_response}")
            resource_dict = create_response.get("cluster", {})

            # Validate resource creation
            if resource_dict is not None:
                self.active_resource = create_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the EcsCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            cluster_name = self.get_ecs_cluster_name()
            describe_response = service_client.describe_clusters(clusters=[cluster_name])
            logger.debug(f"EcsCluster: {describe_response}")
            resource_list = describe_response.get("clusters", None)

            if resource_list is not None and isinstance(resource_list, list):
                for resource in resource_list:
                    _cluster_identifier = resource.get("clusterName", None)
                    if _cluster_identifier == cluster_name:
                        _cluster_status = resource.get("status", None)
                        if _cluster_status == "ACTIVE":
                            self.active_resource = resource
                            break
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EcsCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        self.active_resource = None

        try:
            delete_response = service_client.delete_cluster(cluster=self.get_ecs_cluster_name())
            logger.debug(f"EcsCluster: {delete_response}")
            return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def get_arn(self, aws_client: AwsApiClient) -> Optional[str]:
        tg = self._read(aws_client)
        if tg is None:
            return None
        tg_arn = tg.get("ListenerArn", None)
        return tg_arn



from agno.aws.resource.rds.db_cluster import DbCluster
from agno.aws.resource.rds.db_instance import DbInstance
from agno.aws.resource.rds.db_subnet_group import DbSubnetGroup

__all__ = [
    "DbCluster",
    "DbInstance",
    "DbSubnetGroup",
]



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from typing_extensions import Literal

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.aws.resource.cloudformation.stack import CloudFormationStack
from agno.aws.resource.ec2.security_group import SecurityGroup
from agno.aws.resource.rds.db_subnet_group import DbSubnetGroup
from agno.aws.resource.secret.manager import SecretsManager
from agno.cli.console import print_info
from agno.utilities.logging import logger


class DbInstance(AwsResource):
    """
    The DBInstance can be an RDS DB instance, or it can be a DB instance in an Aurora DB cluster.

    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html
    """

    resource_type: Optional[str] = "DbInstance"
    service_name: str = "rds"

    # Name of the db instance.
    name: str
    # The name of the database engine to be used for this instance.
    engine: Union[
        str,
        Literal[
            "aurora",
            "aurora-mysql",
            "aurora-postgresql",
            "custom-oracle-ee",
            "custom-sqlserver-ee",
            "custom-sqlserver-se",
            "custom-sqlserver-web",
            "mariadb",
            "mysql",
            "oracle-ee",
            "oracle-ee-cdb",
            "oracle-se2",
            "oracle-se2-cdb",
            "postgres",
            "sqlserver-ee",
            "sqlserver-se",
            "sqlserver-ex",
            "sqlserver-web",
        ],
    ]
    # The version number of the database engine to use.
    engine_version: Optional[str] = None
    # Compute and memory capacity of the DB instance, for example db.m5.large.
    db_instance_class: Optional[str] = None

    # This is the name of the database to create when the DB instance is created.
    # Note: The meaning of this parameter differs according to the database engine you use.
    # Provide DB_NAME here or as DB_NAME in secrets_file
    db_name: Optional[str] = None
    # The identifier for this DB instance. This parameter is stored as a lowercase string.
    # If None, use the name as the db_instance_identifier
    # Constraints:
    # - Must contain from 1 to 63 letters, numbers, or hyphens.
    # - First character must be a letter.
    # - Can't end with a hyphen or contain two consecutive hyphens.
    db_instance_identifier: Optional[str] = None
    # The amount of storage in gibibytes (GiB) to allocate for the DB instance.
    allocated_storage: Optional[int] = None
    # The name of the DB parameter group to associate with this DB instance.
    db_parameter_group_name: Optional[str] = None

    # The port number on which the database accepts connections.
    port: Optional[int] = None

    # The name for the master user.
    # Provide MASTER_USERNAME here or as MASTER_USERNAME in secrets_file
    master_username: Optional[str] = None
    # The password for the master user.
    # The password can include any printable ASCII character except "/", """, or "@".
    # Provide MASTER_USER_PASSWORD here or as MASTER_USER_PASSWORD in secrets_file
    master_user_password: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None
    # Read secret variables from AWS Secret
    aws_secret: Optional[SecretsManager] = None

    # The Availability Zone (AZ) where the database will be created.
    availability_zone: Optional[str] = None
    # A DB subnet group to associate with this DB instance.
    db_subnet_group_name: Optional[str] = None
    # If db_subnet_group_name is None,
    # Read the db_subnet_group_name from db_subnet_group
    db_subnet_group: Optional[DbSubnetGroup] = None

    # Specifies whether the DB instance is publicly accessible.
    # When the DB instance is publicly accessible, its Domain Name System (DNS) endpoint resolves to the private IP
    # from within the DB instance's virtual private cloud (VPC). It resolves to the public IP address from outside
    # the DB instance's VPC. Access to the DB instance is ultimately controlled by the security group it uses.
    # That public access is not permitted if the security group assigned to the DB instance doesn't permit it.
    #
    # When the DB instance isn't publicly accessible, it is an internal DB instance with a DNS name that resolves
    # to a private IP address.
    publicly_accessible: Optional[bool] = None
    # The identifier of the DB cluster that the instance will belong to.
    db_cluster_identifier: Optional[str] = None
    # Specifies the storage type to be associated with the DB instance.
    # Valid values: gp2 | gp3 | io1 | standard
    # If you specify io1 or gp3 , you must also include a value for the Iops parameter.
    # Default: io1 if the Iops parameter is specified, otherwise gp2
    storage_type: Optional[str] = None
    iops: Optional[int] = None

    # A list of VPC security groups to associate with this DB instance.
    vpc_security_group_ids: Optional[List[str]] = None
    # If vpc_security_group_ids is None,
    # Read the security_group_id from vpc_stack
    vpc_stack: Optional[CloudFormationStack] = None
    # Add security_group_ids from db_security_groups
    db_security_groups: Optional[List[SecurityGroup]] = None

    backup_retention_period: Optional[int] = None
    character_set_name: Optional[str] = None
    preferred_backup_window: Optional[str] = None
    # The time range each week during which system maintenance can occur, in Universal Coordinated Time (UTC).
    preferred_maintenance_window: Optional[str] = None
    # A value that indicates whether the DB instance is a Multi-AZ deployment.
    # You can't set the AvailabilityZone parameter if the DB instance is a Multi-AZ deployment.
    multi_az: Optional[bool] = None
    auto_minor_version_upgrade: Optional[bool] = None
    license_model: Optional[str] = None
    option_group_name: Optional[str] = None
    nchar_character_set_name: Optional[str] = None
    tags: Optional[List[Dict[str, str]]] = None
    tde_credential_arn: Optional[str] = None
    tde_credential_password: Optional[str] = None
    storage_encrypted: Optional[bool] = None
    kms_key_id: Optional[str] = None
    domain: Optional[str] = None
    copy_tags_to_snapshot: Optional[bool] = None
    monitoring_interval: Optional[int] = None
    monitoring_role_arn: Optional[str] = None
    domain_iam_role_name: Optional[str] = None
    promotion_tier: Optional[int] = None
    timezone: Optional[str] = None
    enable_iam_database_authentication: Optional[bool] = None
    enable_performance_insights: Optional[bool] = None
    performance_insights_kms_key_id: Optional[str] = None
    performance_insights_retention_period: Optional[int] = None
    enable_cloudwatch_logs_exports: Optional[List[str]] = None
    processor_features: Optional[List[Dict[str, str]]] = None
    # A value that indicates whether the DB instance has deletion protection enabled. The database can't be deleted
    # when deletion protection is enabled. By default, deletion protection isn't enabled.
    deletion_protection: Optional[bool] = None
    # The upper limit in gibibytes (GiB) to which Amazon RDS can automatically scale the storage of the DB instance.
    max_allocated_storage: Optional[int] = None
    enable_customer_owned_ip: Optional[bool] = None
    custom_iam_instance_profile: Optional[str] = None
    backup_target: Optional[str] = None
    network_type: Optional[str] = None
    storage_throughput: Optional[int] = None
    ca_certificate_identifier: Optional[str] = None
    db_system_id: Optional[str] = None
    dedicated_log_volume: Optional[bool] = None

    # Specifies whether to manage the master user password with Amazon Web Services Secrets Manager.
    # Constraints:
    # Can’t manage the master user password with Amazon Web Services Secrets Manager if MasterUserPassword is specified.
    manage_master_user_password: Optional[bool] = None
    # The Amazon Web Services KMS key identifier to encrypt a secret that is automatically generated and
    # managed in Amazon Web Services Secrets Manager.
    master_user_secret_kms_key_id: Optional[str] = None

    # Parameters for delete function
    # Skip the creation of a final DB snapshot before the instance is deleted.
    # If skip_final_snapshot = True, no DB snapshot is created.
    # If skip_final_snapshot = None | False, a DB snapshot is created before the instance is deleted.
    #   You must specify a FinalDBSnapshotIdentifier parameter
    #   if skip_final_snapshot = None | False
    skip_final_snapshot: Optional[bool] = True
    # The DB cluster snapshot identifier of the new DB cluster snapshot created when SkipFinalSnapshot is disabled.
    final_db_snapshot_identifier: Optional[str] = None
    # Specifies whether to remove automated backups immediately after the DB cluster is deleted.
    # The default is to remove automated backups immediately after the DB cluster is deleted.
    delete_automated_backups: Optional[bool] = None

    # Parameters for update function
    apply_immediately: Optional[bool] = True
    allow_major_version_upgrade: Optional[bool] = None
    new_db_instance_identifier: Optional[str] = None
    db_port_number: Optional[int] = None
    cloudwatch_logs_export_configuration: Optional[Dict[str, Any]] = None
    use_default_processor_features: Optional[bool] = None
    certificate_rotation_restart: Optional[bool] = None
    replica_mode: Optional[str] = None
    aws_backup_recovery_point_arn: Optional[str] = None
    automation_mode: Optional[str] = None
    resume_full_automation_mode_minutes: Optional[int] = None
    rotate_master_user_password: Optional[bool] = None

    # Cache secret_data
    cached_secret_data: Optional[Dict[str, Any]] = None

    def get_db_instance_identifier(self):
        return self.db_instance_identifier or self.name

    def get_master_username(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        master_username = self.master_username
        if master_username is None and self.secrets_file is not None:
            # read from secrets_file
            secret_data = self.get_secret_file_data()
            if secret_data is not None:
                master_username = secret_data.get("MASTER_USERNAME", master_username)
        if master_username is None and self.aws_secret is not None:
            # read from aws_secret
            logger.debug(f"Reading MASTER_USERNAME from secret: {self.aws_secret.name}")
            master_username = self.aws_secret.get_secret_value("MASTER_USERNAME", aws_client=aws_client)

        return master_username

    def get_master_user_password(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        master_user_password = self.master_user_password
        if master_user_password is None and self.secrets_file is not None:
            # read from secrets_file
            secret_data = self.get_secret_file_data()
            if secret_data is not None:
                master_user_password = secret_data.get("MASTER_USER_PASSWORD", master_user_password)
        if master_user_password is None and self.aws_secret is not None:
            # read from aws_secret
            logger.debug(f"Reading MASTER_USER_PASSWORD from secret: {self.aws_secret.name}")
            master_user_password = self.aws_secret.get_secret_value("MASTER_USER_PASSWORD", aws_client=aws_client)

        return master_user_password

    def get_db_name(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        db_name = self.db_name
        if db_name is None and self.secrets_file is not None:
            # read from secrets_file
            secret_data = self.get_secret_file_data()
            if secret_data is not None:
                db_name = secret_data.get("DB_NAME", db_name)
                if db_name is None:
                    db_name = secret_data.get("DATABASE_NAME", db_name)
        if db_name is None and self.aws_secret is not None:
            # read from aws_secret
            logger.debug(f"Reading DB_NAME from secret: {self.aws_secret.name}")
            db_name = self.aws_secret.get_secret_value("DB_NAME", aws_client=aws_client)
            if db_name is None:
                logger.debug(f"Reading DATABASE_NAME from secret: {self.aws_secret.name}")
                db_name = self.aws_secret.get_secret_value("DATABASE_NAME", aws_client=aws_client)
        return db_name

    def get_database_name(self) -> Optional[str]:
        # Alias for get_db_name because db_instances use `db_name` and db_clusters use `database_name`
        return self.get_db_name()

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the DbInstance

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}

        # Step 1: Get the VpcSecurityGroupIds
        vpc_security_group_ids = self.vpc_security_group_ids
        if vpc_security_group_ids is None and self.vpc_stack is not None:
            vpc_stack_sg = self.vpc_stack.get_security_group(aws_client=aws_client)
            if vpc_stack_sg is not None:
                vpc_security_group_ids = [vpc_stack_sg]
        if self.db_security_groups is not None:
            sg_ids = []
            for sg in self.db_security_groups:
                sg_id = sg.get_security_group_id(aws_client)
                if sg_id is not None:
                    sg_ids.append(sg_id)
            if len(sg_ids) > 0:
                if vpc_security_group_ids is None:
                    vpc_security_group_ids = []
                vpc_security_group_ids.extend(sg_ids)
        if vpc_security_group_ids is not None:
            logger.debug(f"Using SecurityGroups: {vpc_security_group_ids}")
            not_null_args["VpcSecurityGroupIds"] = vpc_security_group_ids

        # Step 2: Get the DbSubnetGroupName
        db_subnet_group_name = self.db_subnet_group_name
        if db_subnet_group_name is None and self.db_subnet_group is not None:
            db_subnet_group_name = self.db_subnet_group.name
            logger.debug(f"Using DbSubnetGroup: {db_subnet_group_name}")
        if db_subnet_group_name is not None:
            not_null_args["DBSubnetGroupName"] = db_subnet_group_name

        db_name = self.get_db_name()
        if db_name:
            not_null_args["DBName"] = db_name

        master_username = self.get_master_username()
        if master_username:
            not_null_args["MasterUsername"] = master_username
        master_user_password = self.get_master_user_password()
        if master_user_password:
            not_null_args["MasterUserPassword"] = master_user_password

        if self.allocated_storage:
            not_null_args["AllocatedStorage"] = self.allocated_storage
        if self.db_instance_class:
            not_null_args["DBInstanceClass"] = self.db_instance_class

        if self.availability_zone is not None:
            not_null_args["AvailabilityZone"] = self.availability_zone

        if self.preferred_maintenance_window:
            not_null_args["PreferredMaintenanceWindow"] = self.preferred_maintenance_window
        if self.db_parameter_group_name:
            not_null_args["DBParameterGroupName"] = self.db_parameter_group_name
        if self.backup_retention_period:
            not_null_args["BackupRetentionPeriod"] = self.backup_retention_period
        if self.preferred_backup_window:
            not_null_args["PreferredBackupWindow"] = self.preferred_backup_window
        if self.port:
            not_null_args["Port"] = self.port
        if self.multi_az:
            not_null_args["MultiAZ"] = self.multi_az
        if self.engine_version:
            not_null_args["EngineVersion"] = self.engine_version
        if self.auto_minor_version_upgrade:
            not_null_args["AutoMinorVersionUpgrade"] = self.auto_minor_version_upgrade
        if self.license_model:
            not_null_args["LicenseModel"] = self.license_model
        if self.iops:
            not_null_args["Iops"] = self.iops
        if self.option_group_name:
            not_null_args["OptionGroupName"] = self.option_group_name
        if self.character_set_name:
            not_null_args["CharacterSetName"] = self.character_set_name
        if self.nchar_character_set_name:
            not_null_args["NcharCharacterSetName"] = self.nchar_character_set_name
        if self.publicly_accessible:
            not_null_args["PubliclyAccessible"] = self.publicly_accessible
        if self.tags:
            not_null_args["Tags"] = self.tags
        if self.db_cluster_identifier:
            not_null_args["DBClusterIdentifier"] = self.db_cluster_identifier
        if self.storage_type:
            not_null_args["StorageType"] = self.storage_type
        if self.tde_credential_arn:
            not_null_args["TdeCredentialArn"] = self.tde_credential_arn
        if self.tde_credential_password:
            not_null_args["TdeCredentialPassword"] = self.tde_credential_password
        if self.storage_encrypted:
            not_null_args["StorageEncrypted"] = self.storage_encrypted
        if self.kms_key_id:
            not_null_args["KmsKeyId"] = self.kms_key_id
        if self.domain:
            not_null_args["Domain"] = self.domain
        if self.copy_tags_to_snapshot:
            not_null_args["CopyTagsToSnapshot"] = self.copy_tags_to_snapshot
        if self.monitoring_interval:
            not_null_args["MonitoringInterval"] = self.monitoring_interval
        if self.monitoring_role_arn:
            not_null_args["MonitoringRoleArn"] = self.monitoring_role_arn
        if self.domain_iam_role_name:
            not_null_args["DomainIAMRoleName"] = self.domain_iam_role_name
        if self.promotion_tier:
            not_null_args["PromotionTier"] = self.promotion_tier
        if self.timezone:
            not_null_args["Timezone"] = self.timezone
        if self.enable_iam_database_authentication:
            not_null_args["EnableIAMDatabaseAuthentication"] = self.enable_iam_database_authentication
        if self.enable_performance_insights:
            not_null_args["EnablePerformanceInsights"] = self.enable_performance_insights
        if self.performance_insights_kms_key_id:
            not_null_args["PerformanceInsightsKMSKeyId"] = self.performance_insights_kms_key_id
        if self.performance_insights_retention_period:
            not_null_args["PerformanceInsightsRetentionPeriod"] = self.performance_insights_retention_period
        if self.enable_cloudwatch_logs_exports:
            not_null_args["EnableCloudwatchLogsExports"] = self.enable_cloudwatch_logs_exports
        if self.processor_features:
            not_null_args["ProcessorFeatures"] = self.processor_features
        if self.deletion_protection:
            not_null_args["DeletionProtection"] = self.deletion_protection
        if self.max_allocated_storage:
            not_null_args["MaxAllocatedStorage"] = self.max_allocated_storage
        if self.enable_customer_owned_ip:
            not_null_args["EnableCustomerOwnedIp"] = self.enable_customer_owned_ip
        if self.custom_iam_instance_profile:
            not_null_args["CustomIamInstanceProfile"] = self.custom_iam_instance_profile
        if self.backup_target:
            not_null_args["BackupTarget"] = self.backup_target
        if self.network_type:
            not_null_args["NetworkType"] = self.network_type
        if self.storage_throughput:
            not_null_args["StorageThroughput"] = self.storage_throughput
        if self.ca_certificate_identifier:
            not_null_args["CACertificateIdentifier"] = self.ca_certificate_identifier
        if self.db_system_id:
            not_null_args["DBSystemId"] = self.db_system_id
        if self.dedicated_log_volume:
            not_null_args["DedicatedLogVolume"] = self.dedicated_log_volume

        if self.manage_master_user_password:
            not_null_args["ManageMasterUserPassword"] = self.manage_master_user_password
        if self.master_user_secret_kms_key_id:
            not_null_args["MasterUserSecretKmsKeyId"] = self.master_user_secret_kms_key_id

        # Step 3: Create DBInstance
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_db_instance(
                DBInstanceIdentifier=self.get_db_instance_identifier(),
                Engine=self.engine,
                **not_null_args,
            )
            logger.debug(f"Response: {create_response}")
            resource_dict = create_response.get("DBInstance", {})

            # Validate resource creation
            if resource_dict is not None:
                logger.debug(f"DBInstance created: {self.get_db_instance_identifier()}")
                self.active_resource = resource_dict
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for DbInstance to be created
        if self.wait_for_create:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be active.")
                waiter = self.get_service_client(aws_client).get_waiter("db_instance_available")
                waiter.wait(
                    DBInstanceIdentifier=self.get_db_instance_identifier(),
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the DbInstance

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            resource_identifier = self.get_db_instance_identifier()
            describe_response = service_client.describe_db_instances(DBInstanceIdentifier=resource_identifier)
            # logger.debug(f"DbInstance: {describe_response}")
            resources_list = describe_response.get("DBInstances", None)

            if resources_list is not None and isinstance(resources_list, list):
                for _resource in resources_list:
                    _identifier = _resource.get("DBInstanceIdentifier", None)
                    if _identifier == resource_identifier:
                        self.active_resource = _resource
                        break
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the DbInstance

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        self.active_resource = None

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.final_db_snapshot_identifier:
            not_null_args["FinalDBSnapshotIdentifier"] = self.final_db_snapshot_identifier
        if self.delete_automated_backups:
            not_null_args["DeleteAutomatedBackups"] = self.delete_automated_backups

        try:
            db_instance_identifier = self.get_db_instance_identifier()
            delete_response = service_client.delete_db_instance(
                DBInstanceIdentifier=db_instance_identifier,
                SkipFinalSnapshot=self.skip_final_snapshot,
                **not_null_args,
            )
            logger.debug(f"Response: {delete_response}")
            resource_dict = delete_response.get("DBInstance", {})

            # Validate resource creation
            if resource_dict is not None:
                logger.debug(f"DBInstance deleted: {self.get_db_instance_identifier()}")
                self.active_resource = resource_dict
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # Wait for DbInstance to be deleted
        if self.wait_for_delete:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be deleted.")
                waiter = self.get_service_client(aws_client).get_waiter("db_instance_deleted")
                waiter.wait(
                    DBInstanceIdentifier=self.get_db_instance_identifier(),
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                logger.error("Waiter failed.")
                logger.error(e)
        return True

    def _update(self, aws_client: AwsApiClient) -> bool:
        """Updates the DbInstance"""

        print_info(f"Updating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Get existing DBInstance
        db_instance = self.read(aws_client)

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {
            "ApplyImmediately": self.apply_immediately,
        }

        vpc_security_group_ids = self.vpc_security_group_ids
        if vpc_security_group_ids is None and self.vpc_stack is not None:
            vpc_stack_sg = self.vpc_stack.get_security_group(aws_client=aws_client)
            if vpc_stack_sg is not None:
                vpc_security_group_ids = [vpc_stack_sg]
        if self.db_security_groups is not None:
            sg_ids = []
            for sg in self.db_security_groups:
                sg_id = sg.get_security_group_id(aws_client)
                if sg_id is not None:
                    sg_ids.append(sg_id)
            if len(sg_ids) > 0:
                if vpc_security_group_ids is None:
                    vpc_security_group_ids = []
                vpc_security_group_ids.extend(sg_ids)
        # Check if vpc_security_group_ids has changed
        existing_vpc_security_group = db_instance.get("VpcSecurityGroups", [])
        existing_vpc_security_group_ids = []
        for existing_sg in existing_vpc_security_group:
            existing_vpc_security_group_ids.append(existing_sg.get("VpcSecurityGroupId", None))
        if vpc_security_group_ids is not None and vpc_security_group_ids != existing_vpc_security_group_ids:
            logger.info(f"Updating SecurityGroups: {vpc_security_group_ids}")
            not_null_args["VpcSecurityGroupIds"] = vpc_security_group_ids

        db_subnet_group_name = self.db_subnet_group_name
        if db_subnet_group_name is None and self.db_subnet_group is not None:
            db_subnet_group_name = self.db_subnet_group.name
        # Check if db_subnet_group_name has changed
        existing_db_subnet_group_name = db_instance.get("DBSubnetGroup", {}).get("DBSubnetGroupName", None)
        if db_subnet_group_name is not None and db_subnet_group_name != existing_db_subnet_group_name:
            logger.info(f"Updating DbSubnetGroup: {db_subnet_group_name}")
            not_null_args["DBSubnetGroupName"] = db_subnet_group_name

        master_user_password = self.get_master_user_password()
        if master_user_password:
            not_null_args["MasterUserPassword"] = master_user_password

        if self.allocated_storage:
            not_null_args["AllocatedStorage"] = self.allocated_storage
        if self.db_instance_class:
            not_null_args["DBInstanceClass"] = self.db_instance_class

        if self.db_parameter_group_name:
            not_null_args["DBParameterGroupName"] = self.db_parameter_group_name
        if self.backup_retention_period:
            not_null_args["BackupRetentionPeriod"] = self.backup_retention_period
        if self.preferred_backup_window:
            not_null_args["PreferredBackupWindow"] = self.preferred_backup_window
        if self.preferred_maintenance_window:
            not_null_args["PreferredMaintenanceWindow"] = self.preferred_maintenance_window
        if self.multi_az:
            not_null_args["MultiAZ"] = self.multi_az
        if self.engine_version:
            not_null_args["EngineVersion"] = self.engine_version
        if self.allow_major_version_upgrade:
            not_null_args["AllowMajorVersionUpgrade"] = self.allow_major_version_upgrade
        if self.auto_minor_version_upgrade:
            not_null_args["AutoMinorVersionUpgrade"] = self.auto_minor_version_upgrade
        if self.license_model:
            not_null_args["LicenseModel"] = self.license_model
        if self.iops:
            not_null_args["Iops"] = self.iops
        if self.option_group_name:
            not_null_args["OptionGroupName"] = self.option_group_name
        if self.new_db_instance_identifier:
            not_null_args["NewDBInstanceIdentifier"] = self.new_db_instance_identifier
        if self.storage_type:
            not_null_args["StorageType"] = self.storage_type
        if self.tde_credential_arn:
            not_null_args["TdeCredentialArn"] = self.tde_credential_arn
        if self.tde_credential_password:
            not_null_args["TdeCredentialPassword"] = self.tde_credential_password
        if self.ca_certificate_identifier:
            not_null_args["CACertificateIdentifier"] = self.ca_certificate_identifier
        if self.domain:
            not_null_args["Domain"] = self.domain
        if self.copy_tags_to_snapshot:
            not_null_args["CopyTagsToSnapshot"] = self.copy_tags_to_snapshot
        if self.monitoring_interval:
            not_null_args["MonitoringInterval"] = self.monitoring_interval
        if self.db_port_number:
            not_null_args["DBPortNumber"] = self.db_port_number
        if self.publicly_accessible:
            not_null_args["PubliclyAccessible"] = self.publicly_accessible
        if self.monitoring_role_arn:
            not_null_args["MonitoringRoleArn"] = self.monitoring_role_arn
        if self.domain_iam_role_name:
            not_null_args["DomainIAMRoleName"] = self.domain_iam_role_name
        if self.promotion_tier:
            not_null_args["PromotionTier"] = self.promotion_tier
        if self.enable_iam_database_authentication:
            not_null_args["EnableIAMDatabaseAuthentication"] = self.enable_iam_database_authentication
        if self.enable_performance_insights:
            not_null_args["EnablePerformanceInsights"] = self.enable_performance_insights
        if self.performance_insights_kms_key_id:
            not_null_args["PerformanceInsightsKMSKeyId"] = self.performance_insights_kms_key_id
        if self.performance_insights_retention_period:
            not_null_args["PerformanceInsightsRetentionPeriod"] = self.performance_insights_retention_period
        if self.cloudwatch_logs_export_configuration:
            not_null_args["CloudwatchLogsExportConfiguration"] = self.cloudwatch_logs_export_configuration
        if self.processor_features:
            not_null_args["ProcessorFeatures"] = self.processor_features
        if self.use_default_processor_features:
            not_null_args["UseDefaultProcessorFeatures"] = self.use_default_processor_features
        if self.deletion_protection:
            not_null_args["DeletionProtection"] = self.deletion_protection
        if self.max_allocated_storage:
            not_null_args["MaxAllocatedStorage"] = self.max_allocated_storage
        if self.certificate_rotation_restart:
            not_null_args["CertificateRotationRestart"] = self.certificate_rotation_restart
        if self.replica_mode:
            not_null_args["ReplicaMode"] = self.replica_mode
        if self.enable_customer_owned_ip:
            not_null_args["EnableCustomerOwnedIp"] = self.enable_customer_owned_ip
        if self.aws_backup_recovery_point_arn:
            not_null_args["AwsBackupRecoveryPointArn"] = self.aws_backup_recovery_point_arn
        if self.automation_mode:
            not_null_args["AutomationMode"] = self.automation_mode
        if self.resume_full_automation_mode_minutes:
            not_null_args["ResumeFullAutomationModeMinutes"] = self.resume_full_automation_mode_minutes
        if self.network_type:
            not_null_args["NetworkType"] = self.network_type
        if self.storage_throughput:
            not_null_args["StorageThroughput"] = self.storage_throughput
        if self.manage_master_user_password:
            not_null_args["ManageMasterUserPassword"] = self.manage_master_user_password
        if self.rotate_master_user_password:
            not_null_args["RotateMasterUserPassword"] = self.rotate_master_user_password
        if self.master_user_secret_kms_key_id:
            not_null_args["MasterUserSecretKmsKeyId"] = self.master_user_secret_kms_key_id

        # Step 2: Update DBInstance
        service_client = self.get_service_client(aws_client)
        try:
            update_response = service_client.modify_db_instance(
                DBInstanceIdentifier=self.get_db_instance_identifier(),
                **not_null_args,
            )
            logger.debug(f"Response: {update_response}")
            resource_dict = update_response.get("DBInstance", {})

            # Validate resource update
            if resource_dict is not None:
                print_info(f"DBInstance updated: {self.get_resource_name()}")
                self.active_resource = update_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def get_db_endpoint(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        """Returns the DbInstance endpoint

        Returns:
            The DbInstance endpoint
        """
        logger.debug(f"Getting endpoint for {self.get_resource_name()}")
        _db_endpoint: Optional[str] = None
        if self.active_resource:
            _db_endpoint = self.active_resource.get("Endpoint", {}).get("Address", None)
        if _db_endpoint is None:
            client: AwsApiClient = aws_client or self.get_aws_client()
            resource = self._read(aws_client=client)
            if resource is not None:
                _db_endpoint = resource.get("Endpoint", {}).get("Address", None)
        if _db_endpoint is None:
            resource = self.read_resource_from_file()
            if resource is not None:
                _db_endpoint = resource.get("Endpoint", {}).get("Address", None)
        logger.debug(f"DBInstance Endpoint: {_db_endpoint}")
        return _db_endpoint

    def get_db_port(self, aws_client: Optional[AwsApiClient] = None) -> Optional[str]:
        """Returns the DbInstance port

        Returns:
            The DbInstance port
        """
        logger.debug(f"Getting port for {self.get_resource_name()}")
        _db_port: Optional[str] = None
        if self.active_resource:
            _db_port = self.active_resource.get("Endpoint", {}).get("Port", None)
        if _db_port is None:
            client: AwsApiClient = aws_client or self.get_aws_client()
            resource = self._read(aws_client=client)
            if resource is not None:
                _db_port = resource.get("Endpoint", {}).get("Port", None)
        if _db_port is None:
            resource = self.read_resource_from_file()
            if resource is not None:
                _db_port = resource.get("Endpoint", {}).get("Port", None)
        logger.debug(f"DBInstance Port: {_db_port}")
        return _db_port



from typing import Dict, List, Optional, Union

from agno.docker.app.base import ContainerContext, DockerApp  # noqa: F401


class Streamlit(DockerApp):
    # -*- App Name
    name: str = "streamlit"

    # -*- Image Configuration
    image_name: str = "agnohq/streamlit"
    image_tag: str = "1.27"
    command: Optional[Union[str, List[str]]] = "streamlit hello"

    # -*- App Ports
    # Open a container port if open_port=True
    open_port: bool = True
    port_number: int = 8501

    # -*- OS Configuration
    # Path to the os directory inside the container
    infra_dir_container_path: str = "/app"
    # Mount the infra directory from host machine to the container
    mount_infra: bool = False

    # -*- Streamlit Configuration
    # Server settings
    # Defaults to the port_number
    streamlit_server_port: Optional[int] = None
    streamlit_server_headless: bool = True
    streamlit_server_run_on_save: Optional[bool] = None
    streamlit_server_max_upload_size: Optional[int] = None
    streamlit_browser_gather_usage_stats: bool = False
    # Browser settings
    streamlit_browser_server_port: Optional[str] = None
    streamlit_browser_server_address: Optional[str] = None

    def get_container_env(self, container_context: ContainerContext) -> Dict[str, str]:
        container_env: Dict[str, str] = super().get_container_env(container_context=container_context)

        streamlit_server_port = self.streamlit_server_port
        if streamlit_server_port is None:
            port_number = self.port_number
            if port_number is not None:
                streamlit_server_port = port_number
        if streamlit_server_port is not None:
            container_env["STREAMLIT_SERVER_PORT"] = str(streamlit_server_port)

        if self.streamlit_server_headless is not None:
            container_env["STREAMLIT_SERVER_HEADLESS"] = str(self.streamlit_server_headless)

        if self.streamlit_server_run_on_save is not None:
            container_env["STREAMLIT_SERVER_RUN_ON_SAVE"] = str(self.streamlit_server_run_on_save)

        if self.streamlit_server_max_upload_size is not None:
            container_env["STREAMLIT_SERVER_MAX_UPLOAD_SIZE"] = str(self.streamlit_server_max_upload_size)

        if self.streamlit_browser_gather_usage_stats is not None:
            container_env["STREAMLIT_BROWSER_GATHER_USAGE_STATS"] = str(self.streamlit_browser_gather_usage_stats)

        if self.streamlit_browser_server_port is not None:
            container_env["STREAMLIT_BROWSER_SERVER_PORT"] = self.streamlit_browser_server_port

        if self.streamlit_browser_server_address is not None:
            container_env["STREAMLIT_BROWSER_SERVER_ADDRESS"] = self.streamlit_browser_server_address

        return container_env



from typing import Any, Dict, List, Optional, Union

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.base import AwsResource
from agno.aws.resource.ec2.subnet import Subnet
from agno.cli.console import print_info
from agno.utilities.logging import logger


class TargetGroup(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/elbv2/client/create_target_group.html
    """

    resource_type: Optional[str] = "TargetGroup"
    service_name: str = "elbv2"

    # Name of the Target Group
    name: str
    protocol: Optional[str] = None
    protocol_version: Optional[str] = None
    port: Optional[int] = None
    vpc_id: Optional[str] = None
    subnets: Optional[List[Union[str, Subnet]]] = None
    health_check_protocol: Optional[str] = None
    health_check_port: Optional[str] = None
    health_check_enabled: Optional[bool] = None
    health_check_path: Optional[str] = None
    health_check_interval_seconds: Optional[int] = None
    health_check_timeout_seconds: Optional[int] = None
    healthy_threshold_count: Optional[int] = None
    unhealthy_threshold_count: Optional[int] = None
    matcher: Optional[Dict[str, str]] = None
    target_type: Optional[str] = None
    tags: Optional[List[Dict[str, str]]] = None
    ip_address_type: Optional[str] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the Target Group

        Args:
            aws_client: The AwsApiClient for the current Target Group
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}

        # Get vpc_id
        vpc_id = self.vpc_id
        if vpc_id is None and self.subnets is not None:
            from agno.aws.resource.ec2.subnet import get_vpc_id_from_subnet_ids

            subnet_ids = []
            for subnet in self.subnets:
                if isinstance(subnet, Subnet):
                    subnet_ids.append(subnet.name)
                elif isinstance(subnet, str):
                    subnet_ids.append(subnet)
            vpc_id = get_vpc_id_from_subnet_ids(subnet_ids, aws_client)
        if vpc_id is not None:
            not_null_args["VpcId"] = vpc_id

        if self.protocol is not None:
            not_null_args["Protocol"] = self.protocol
        if self.protocol_version is not None:
            not_null_args["ProtocolVersion"] = self.protocol_version
        if self.port is not None:
            not_null_args["Port"] = self.port
        if self.health_check_protocol is not None:
            not_null_args["HealthCheckProtocol"] = self.health_check_protocol
        if self.health_check_port is not None:
            not_null_args["HealthCheckPort"] = self.health_check_port
        if self.health_check_enabled is not None:
            not_null_args["HealthCheckEnabled"] = self.health_check_enabled
        if self.health_check_path is not None:
            not_null_args["HealthCheckPath"] = self.health_check_path
        if self.health_check_interval_seconds is not None:
            not_null_args["HealthCheckIntervalSeconds"] = self.health_check_interval_seconds
        if self.health_check_timeout_seconds is not None:
            not_null_args["HealthCheckTimeoutSeconds"] = self.health_check_timeout_seconds
        if self.healthy_threshold_count is not None:
            not_null_args["HealthyThresholdCount"] = self.healthy_threshold_count
        if self.unhealthy_threshold_count is not None:
            not_null_args["UnhealthyThresholdCount"] = self.unhealthy_threshold_count
        if self.matcher is not None:
            not_null_args["Matcher"] = self.matcher
        if self.target_type is not None:
            not_null_args["TargetType"] = self.target_type
        if self.tags is not None:
            not_null_args["Tags"] = self.tags
        if self.ip_address_type is not None:
            not_null_args["IpAddressType"] = self.ip_address_type

        # Create TargetGroup
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_target_group(
                Name=self.name,
                **not_null_args,
            )
            logger.debug(f"Response: {create_response}")
            resource_dict = create_response.get("TargetGroups", {})

            # Validate resource creation
            if resource_dict is not None:
                self.active_resource = create_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the TargetGroup

        Args:
            aws_client: The AwsApiClient for the current TargetGroup
        """
        from botocore.exceptions import ClientError

        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        service_client = self.get_service_client(aws_client)
        try:
            describe_response = service_client.describe_target_groups(Names=[self.name])
            logger.debug(f"Describe Response: {describe_response}")
            resource_list = describe_response.get("TargetGroups", None)

            if resource_list is not None and isinstance(resource_list, list):
                for resource in resource_list:
                    if resource.get("TargetGroupName") == self.name:
                        self.active_resource = resource
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the TargetGroup

        Args:
            aws_client: The AwsApiClient for the current TargetGroup
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        self.active_resource = None

        try:
            tg_arn = self.get_arn(aws_client)
            if tg_arn is None:
                logger.error(f"TargetGroup {self.get_resource_name()} not found.")
                return True
            delete_response = service_client.delete_target_group(TargetGroupArn=tg_arn)
            logger.debug(f"Delete Response: {delete_response}")
            return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def _update(self, aws_client: AwsApiClient) -> bool:
        """Update EcsService"""
        print_info(f"Updating {self.get_resource_type()}: {self.get_resource_name()}")

        tg_arn = self.get_arn(aws_client=aws_client)
        if tg_arn is None:
            logger.error(f"TargetGroup {self.get_resource_name()} not found.")
            return True

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.health_check_protocol is not None:
            not_null_args["HealthCheckProtocol"] = self.health_check_protocol
        if self.health_check_port is not None:
            not_null_args["HealthCheckPort"] = self.health_check_port
        if self.health_check_enabled is not None:
            not_null_args["HealthCheckEnabled"] = self.health_check_enabled
        if self.health_check_path is not None:
            not_null_args["HealthCheckPath"] = self.health_check_path
        if self.health_check_interval_seconds is not None:
            not_null_args["HealthCheckIntervalSeconds"] = self.health_check_interval_seconds
        if self.health_check_timeout_seconds is not None:
            not_null_args["HealthCheckTimeoutSeconds"] = self.health_check_timeout_seconds
        if self.healthy_threshold_count is not None:
            not_null_args["HealthyThresholdCount"] = self.healthy_threshold_count
        if self.unhealthy_threshold_count is not None:
            not_null_args["UnhealthyThresholdCount"] = self.unhealthy_threshold_count
        if self.matcher is not None:
            not_null_args["Matcher"] = self.matcher

        service_client = self.get_service_client(aws_client)
        try:
            response = service_client.modify_target_group(
                TargetGroupArn=tg_arn,
                **not_null_args,
            )
            logger.debug(f"Update Response: {response}")
            resource_dict = response.get("TargetGroups", {})

            # Validate resource creation
            if resource_dict is not None:
                print_info(f"TargetGroup updated: {self.get_resource_name()}")
                self.active_resource = response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def get_arn(self, aws_client: AwsApiClient) -> Optional[str]:
        tg = self._read(aws_client)
        if tg is None:
            return None
        tg_arn = tg.get("TargetGroupArn", None)
        return tg_arn



from pathlib import Path
from typing import List, Optional

from agno.base.resources import InfraResources
from agno.cli.config import AgnoCliConfig
from agno.cli.console import print_heading, print_info
from agno.cli.settings import AGNO_CLI_CONFIG_DIR
from agno.utilities.logging import logger


def delete_agno_config() -> None:
    from agno.utilities.filesystem import delete_from_fs

    logger.debug("Removing existing Agno configuration")
    delete_from_fs(AGNO_CLI_CONFIG_DIR)


def initialize_agno_cli(reset: bool = False) -> Optional[AgnoCliConfig]:
    """Initialize Agno CLI config.

    Steps:
    1. Check if AGNO_CLI_CONFIG_DIR exists, if not, create it. If reset == True, recreate AGNO_CLI_CONFIG_DIR.
    2. If AgnoCliConfig exists and auth is valid, returns AgnoCliConfig.
    """
    from agno.utilities.filesystem import delete_from_fs

    print_heading("Welcome to Agno!")
    if reset:
        delete_agno_config()

    logger.debug("Initializing Agno")

    # Check if ~/.config/ag exists, if it is not a dir - delete it and create the directory
    if AGNO_CLI_CONFIG_DIR.exists():
        logger.debug(f"{AGNO_CLI_CONFIG_DIR} exists")
        if not AGNO_CLI_CONFIG_DIR.is_dir():
            try:
                delete_from_fs(AGNO_CLI_CONFIG_DIR)
            except Exception as e:
                logger.exception(e)
                raise Exception(f"Something went wrong, please delete {AGNO_CLI_CONFIG_DIR} and run again")
            AGNO_CLI_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    else:
        AGNO_CLI_CONFIG_DIR.mkdir(parents=True)
        logger.debug(f"Created {AGNO_CLI_CONFIG_DIR}")

    # Confirm AGNO_CLI_CONFIG_DIR exists otherwise we should return
    if AGNO_CLI_CONFIG_DIR.exists():
        logger.debug(f"Agno config location: {AGNO_CLI_CONFIG_DIR}")
    else:
        raise Exception("Something went wrong, please try again")

    agno_config: Optional[AgnoCliConfig] = AgnoCliConfig.from_saved_config()
    if agno_config is None:
        logger.debug("Creating new AgnoCliConfig")
        agno_config = AgnoCliConfig()
        agno_config.save_config()

    logger.debug("Agno CLI initialized")
    return agno_config


def start_resources(
    resources_file_path: Path,
    target_env: Optional[str] = None,
    target_infra: Optional[str] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
    pull: Optional[bool] = False,
) -> None:
    print_heading(f"Starting resources in: {resources_file_path}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")
    logger.debug(f"\tpull         : {pull}")

    from agno.infra.config import InfraConfig

    if not resources_file_path.exists():
        logger.error(f"File does not exist: {resources_file_path}")
        return

    # Get resources to deploy
    resource_groups_to_create: List[InfraResources] = InfraConfig.get_resources_from_file(
        resource_file=resources_file_path,
        env=target_env,
        infra=target_infra,
        order="create",
    )

    # Track number of resource groups created
    num_rgs_created = 0
    num_rgs_to_create = len(resource_groups_to_create)
    # Track number of resources created
    num_resources_created = 0
    num_resources_to_create = 0

    if num_rgs_to_create == 0:
        print_info("No resources to create")
        return

    logger.debug(f"Deploying {num_rgs_to_create} resource groups")
    for rg in resource_groups_to_create:
        _num_resources_created, _num_resources_to_create = rg.create_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
            pull=pull,
        )
        if _num_resources_created > 0:
            num_rgs_created += 1
        num_resources_created += _num_resources_created
        num_resources_to_create += _num_resources_to_create
        logger.debug(f"Deployed {num_resources_created} resources in {num_rgs_created} resource groups")

    if dry_run:
        return

    if num_resources_created == 0:
        return

    print_heading(f"\n--**-- ResourceGroups deployed: {num_rgs_created}/{num_rgs_to_create}\n")
    if num_resources_created != num_resources_to_create:
        logger.error("Some resources failed to create, please check logs")


def stop_resources(
    resources_file_path: Path,
    target_env: Optional[str] = None,
    target_infra: Optional[str] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    print_heading(f"Stopping resources in: {resources_file_path}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")

    from agno.infra.config import InfraConfig

    if not resources_file_path.exists():
        logger.error(f"File does not exist: {resources_file_path}")
        return

    # Get resource groups to shutdown
    resource_groups_to_shutdown: List[InfraResources] = InfraConfig.get_resources_from_file(
        resource_file=resources_file_path,
        env=target_env,
        infra=target_infra,
        order="create",
    )

    # Track number of resource groups deleted
    num_rgs_shutdown = 0
    num_rgs_to_shutdown = len(resource_groups_to_shutdown)
    # Track number of resources created
    num_resources_shutdown = 0
    num_resources_to_shutdown = 0

    if num_rgs_to_shutdown == 0:
        print_info("No resources to delete")
        return

    logger.debug(f"Deleting {num_rgs_to_shutdown} resource groups")
    for rg in resource_groups_to_shutdown:
        _num_resources_shutdown, _num_resources_to_shutdown = rg.delete_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_shutdown > 0:
            num_rgs_shutdown += 1
        num_resources_shutdown += _num_resources_shutdown
        num_resources_to_shutdown += _num_resources_to_shutdown
        logger.debug(f"Deleted {num_resources_shutdown} resources in {num_rgs_shutdown} resource groups")

    if dry_run:
        return

    if num_resources_shutdown == 0:
        return

    print_heading(f"\n--**-- ResourceGroups deleted: {num_rgs_shutdown}/{num_rgs_to_shutdown}\n")
    if num_resources_shutdown != num_resources_to_shutdown:
        logger.error("Some resources failed to delete, please check logs")


def patch_resources(
    resources_file_path: Path,
    target_env: Optional[str] = None,
    target_infra: Optional[str] = None,
    target_group: Optional[str] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
    auto_confirm: Optional[bool] = False,
    force: Optional[bool] = None,
) -> None:
    print_heading(f"Updating resources in: {resources_file_path}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_infra : {target_infra}")
    logger.debug(f"\ttarget_name  : {target_name}")
    logger.debug(f"\ttarget_type  : {target_type}")
    logger.debug(f"\ttarget_group : {target_group}")
    logger.debug(f"\tdry_run      : {dry_run}")
    logger.debug(f"\tauto_confirm : {auto_confirm}")
    logger.debug(f"\tforce        : {force}")

    from agno.infra.config import InfraConfig

    if not resources_file_path.exists():
        logger.error(f"File does not exist: {resources_file_path}")
        return

    # Get resource groups to update
    resource_groups_to_patch: List[InfraResources] = InfraConfig.get_resources_from_file(
        resource_file=resources_file_path,
        env=target_env,
        infra=target_infra,
        order="create",
    )

    num_rgs_patched = 0
    num_rgs_to_patch = len(resource_groups_to_patch)
    # Track number of resources updated
    num_resources_patched = 0
    num_resources_to_patch = 0

    if num_rgs_to_patch == 0:
        print_info("No resources to patch")
        return

    logger.debug(f"Patching {num_rgs_to_patch} resource groups")
    for rg in resource_groups_to_patch:
        _num_resources_patched, _num_resources_to_patch = rg.update_resources(
            group_filter=target_group,
            name_filter=target_name,
            type_filter=target_type,
            dry_run=dry_run,
            auto_confirm=auto_confirm,
            force=force,
        )
        if _num_resources_patched > 0:
            num_rgs_patched += 1
        num_resources_patched += _num_resources_patched
        num_resources_to_patch += _num_resources_to_patch
        logger.debug(f"Patched {num_resources_patched} resources in {num_rgs_patched} resource groups")

    if dry_run:
        return

    if num_resources_patched == 0:
        return

    print_heading(f"\n--**-- ResourceGroups patched: {num_rgs_patched}/{num_rgs_to_patch}\n")
    if num_resources_patched != num_resources_to_patch:
        logger.error("Some resources failed to patch, please check logs")



from typing import Any, Dict, List, Optional

from agno.aws.api_client import AwsApiClient
from agno.aws.resource.acm.certificate import AcmCertificate
from agno.aws.resource.base import AwsResource
from agno.aws.resource.elb.load_balancer import LoadBalancer
from agno.aws.resource.elb.target_group import TargetGroup
from agno.cli.console import print_info
from agno.utilities.logging import logger


class Listener(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/elbv2/client/create_listener.html
    """

    resource_type: Optional[str] = "Listener"
    service_name: str = "elbv2"

    # Name of the Listener
    name: str
    load_balancer: Optional[LoadBalancer] = None
    target_group: Optional[TargetGroup] = None
    load_balancer_arn: Optional[str] = None
    protocol: Optional[str] = None
    port: Optional[int] = None
    ssl_policy: Optional[str] = None
    certificates: Optional[List[Dict[str, Any]]] = None
    acm_certificates: Optional[List[AcmCertificate]] = None
    default_actions: Optional[List[Dict]] = None
    alpn_policy: Optional[List[str]] = None
    tags: Optional[List[Dict[str, str]]] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the Listener

        Args:
            aws_client: The AwsApiClient for the current Listener
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        load_balancer_arn = self.get_load_balancer_arn(aws_client)
        if load_balancer_arn is None:
            logger.error("Load balancer ARN not available")
            return False

        listener_port = self.get_listener_port()
        listener_protocol = self.get_listener_protocol()
        listener_certificates = self.get_listener_certificates(aws_client)

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if listener_port is not None:
            not_null_args["Port"] = listener_port
        if listener_protocol is not None:
            not_null_args["Protocol"] = listener_protocol
        if listener_certificates is not None:
            not_null_args["Certificates"] = listener_certificates
        if self.ssl_policy is not None:
            not_null_args["SslPolicy"] = self.ssl_policy
        if self.alpn_policy is not None:
            not_null_args["AlpnPolicy"] = self.alpn_policy

        # listener tags container a name for the listener
        listener_tags = self.get_listener_tags()
        if listener_tags is not None:
            not_null_args["Tags"] = listener_tags

        if self.default_actions is not None:
            not_null_args["DefaultActions"] = self.default_actions
        elif self.target_group is not None:
            target_group_arn = self.target_group.get_arn(aws_client)
            if target_group_arn is None:
                logger.error("Target group ARN not available")
                return False
            not_null_args["DefaultActions"] = [{"Type": "forward", "TargetGroupArn": target_group_arn}]
        else:
            logger.warning(f"Neither target group nor default actions provided for {self.get_resource_name()}")
            return True

        # Create Listener
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_listener(
                LoadBalancerArn=load_balancer_arn,
                **not_null_args,
            )
            logger.debug(f"Create Response: {create_response}")
            resource_dict = create_response.get("Listeners", {})

            # Validate resource creation
            if resource_dict is not None:
                self.active_resource = create_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the Listener

        Args:
            aws_client: The AwsApiClient for the current Listener
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            load_balancer_arn = self.get_load_balancer_arn(aws_client)
            if load_balancer_arn is None:
                # logger.error(f"Load balancer ARN not available")
                return None

            describe_response = service_client.describe_listeners(LoadBalancerArn=load_balancer_arn)
            logger.debug(f"Describe Response: {describe_response}")
            resource_list = describe_response.get("Listeners", None)

            if resource_list is not None and isinstance(resource_list, list):
                # We identify the current listener by the port and protocol
                current_listener_port = self.get_listener_port()
                current_listener_protocol = self.get_listener_protocol()
                for resource in resource_list:
                    if (
                        resource.get("Port", None) == current_listener_port
                        and resource.get("Protocol", None) == current_listener_protocol
                    ):
                        logger.debug(f"Found {self.get_resource_type()}: {self.get_resource_name()}")
                        self.active_resource = resource
                        break
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the Listener

        Args:
            aws_client: The AwsApiClient for the current Listener
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        self.active_resource = None

        try:
            listener_arn = self.get_arn(aws_client)
            if listener_arn is None:
                logger.error(f"Listener {self.get_resource_name()} not found.")
                return True

            delete_response = service_client.delete_listener(ListenerArn=listener_arn)
            logger.debug(f"Delete Response: {delete_response}")
            return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def _update(self, aws_client: AwsApiClient) -> bool:
        """Update EcsService"""
        print_info(f"Updating {self.get_resource_type()}: {self.get_resource_name()}")

        listener_arn = self.get_arn(aws_client)
        if listener_arn is None:
            logger.error(f"Listener {self.get_resource_name()} not found.")
            return True

        listener_port = self.get_listener_port()
        listener_protocol = self.get_listener_protocol()
        listener_certificates = self.get_listener_certificates(aws_client)

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if listener_port is not None:
            not_null_args["Port"] = listener_port
        if listener_protocol is not None:
            not_null_args["Protocol"] = listener_protocol
        if listener_certificates is not None:
            not_null_args["Certificates"] = listener_certificates
        if self.ssl_policy is not None:
            not_null_args["SslPolicy"] = self.ssl_policy
        if self.alpn_policy is not None:
            not_null_args["AlpnPolicy"] = self.alpn_policy

        if self.default_actions is not None:
            not_null_args["DefaultActions"] = self.default_actions
        elif self.target_group is not None:
            target_group_arn = self.target_group.get_arn(aws_client)
            if target_group_arn is None:
                logger.error("Target group ARN not available")
                return False
            not_null_args["DefaultActions"] = [{"Type": "forward", "TargetGroupArn": target_group_arn}]
        else:
            logger.warning(f"Neither target group nor default actions provided for {self.get_resource_name()}")
            return True

        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.modify_listener(
                ListenerArn=listener_arn,
                **not_null_args,
            )
            logger.debug(f"Update Response: {create_response}")
            resource_dict = create_response.get("Listeners", {})

            # Validate resource creation
            if resource_dict is not None:
                print_info(f"Listener updated: {self.get_resource_name()}")
                self.active_resource = create_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def get_arn(self, aws_client: AwsApiClient) -> Optional[str]:
        listener = self._read(aws_client)
        if listener is None:
            return None

        listener_arn = listener.get("ListenerArn", None)
        return listener_arn

    def get_load_balancer_arn(self, aws_client: AwsApiClient):
        load_balancer_arn = self.load_balancer_arn
        if load_balancer_arn is None and self.load_balancer:
            load_balancer_arn = self.load_balancer.get_arn(aws_client)

        return load_balancer_arn

    def get_listener_port(self):
        listener_port = self.port
        if listener_port is None and self.load_balancer:
            lb_protocol = self.load_balancer.protocol
            listener_port = 443 if lb_protocol == "HTTPS" else 80

        return listener_port

    def get_listener_protocol(self):
        listener_protocol = self.protocol
        if listener_protocol is None and self.load_balancer:
            listener_protocol = self.load_balancer.protocol

        return listener_protocol

    def get_listener_certificates(self, aws_client: AwsApiClient):
        listener_protocol = self.protocol
        if listener_protocol is None and self.load_balancer:
            listener_protocol = self.load_balancer.protocol

        certificates = self.certificates
        if certificates is None and self.acm_certificates is not None and len(self.acm_certificates) > 0:
            certificates = []
            for cert in self.acm_certificates:
                certificates.append({"CertificateArn": cert.get_certificate_arn(aws_client)})

        return certificates

    def get_listener_tags(self):
        tags = self.tags
        if tags is None:
            tags = []
        tags.append({"Key": "Name", "Value": self.get_resource_name()})

        return tags



from typing import Optional

from pydantic import BaseModel


class ContainerContext(BaseModel):
    """ContainerContext is a context object passed when creating containers."""

    # Infra name
    infra_name: str
    # Path to the infra directory inside the container
    infra_root: str
    # Path to the infra parent directory inside the container
    infra_parent: str
    # Path to the requirements.txt file relative to the infra_root
    requirements_file: Optional[str] = None



from typing import List, Optional, Union

from agno.aws.app.base import AwsApp, AwsBuildContext, ContainerContext  # noqa: F401


class CeleryWorker(AwsApp):
    # -*- App Name
    name: str = "celery-worker"

    # -*- Image Configuration
    image_name: str = "agnohq/celery-worker"
    image_tag: str = "latest"
    command: Optional[Union[str, List[str]]] = "celery -A tasks.celery worker --loglevel=info"

    # -*- OS Configuration
    # Path to the os directory inside the container
    infra_dir_container_path: str = "/app"



from typing import Any, Dict, List, Optional

from agno.cli.console import console, print_info
from agno.docker.api_client import DockerApiClient
from agno.docker.resource.base import DockerResource
from agno.utilities.logging import logger


class DockerImage(DockerResource):
    resource_type: str = "Image"

    # Docker image name, usually as repo/image
    name: str
    # Docker image tag
    tag: Optional[str] = None

    # Path to the directory containing the Dockerfile
    path: Optional[str] = None
    # Path to the Dockerfile within the build context
    dockerfile: Optional[str] = None

    # Print the build log
    print_build_log: bool = True
    # Push the image to the registry. Similar to the docker push command.
    push_image: bool = False
    print_push_output: bool = False
    # Use buildx for building images
    use_buildx: bool = True

    # Remove intermediate containers.
    # The docker build command defaults to --rm=true,
    # The docker api kept the old default of False to preserve backward compatibility
    rm: Optional[bool] = True
    # Always remove intermediate containers, even after unsuccessful builds
    forcerm: Optional[bool] = None
    # HTTP timeout
    timeout: Optional[int] = None
    # Downloads any updates to the FROM image in Dockerfiles
    pull: Optional[bool] = None
    # Skips docker cache when set to True
    # i.e. rebuilds all layes of the image
    skip_docker_cache: Optional[bool] = None
    # A dictionary of build arguments
    buildargs: Optional[Dict[str, Any]] = None
    # A dictionary of limits applied to each container created by the build process. Valid keys:
    # memory (int): set memory limit for build
    # memswap (int): Total memory (memory + swap), -1 to disable swap
    # cpushares (int): CPU shares (relative weight)
    # cpusetcpus (str): CPUs in which to allow execution, e.g. "0-3", "0,1"
    container_limits: Optional[Dict[str, Any]] = None
    # Size of /dev/shm in bytes. The size must be greater than 0. If omitted the system uses 64MB
    shmsize: Optional[int] = None
    # A dictionary of labels to set on the image
    labels: Optional[Dict[str, Any]] = None
    # A list of images used for build cache resolution
    cache_from: Optional[List[Any]] = None
    # Name of the build-stage to build in a multi-stage Dockerfile
    target: Optional[str] = None
    # networking mode for the run commands during build
    network_mode: Optional[str] = None
    # Squash the resulting images layers into a single layer.
    squash: Optional[bool] = None
    # Extra hosts to add to /etc/hosts in building containers, as a mapping of hostname to IP address.
    extra_hosts: Optional[Dict[str, Any]] = None
    # Platform in the format os[/arch[/variant]].
    platform: Optional[str] = None
    # List of platforms to use for build, uses buildx_image if multi-platform build is enabled.
    platforms: Optional[List[str]] = None
    # Isolation technology used during build. Default: None.
    isolation: Optional[str] = None
    # If True, and if the docker client configuration file (~/.docker/config.json by default)
    # contains a proxy configuration, the corresponding environment variables
    # will be set in the container being built.
    use_config_proxy: Optional[bool] = None

    # Set skip_delete=True so that the image is not deleted when the `ag ws down` command is run
    skip_delete: bool = True
    image_build_id: Optional[str] = None

    # Set use_cache to False so image is always built
    use_cache: bool = False

    def get_image_str(self) -> str:
        if self.tag:
            return f"{self.name}:{self.tag}"
        return f"{self.name}:latest"

    def get_resource_name(self) -> str:
        return self.get_image_str()

    def buildx(self, docker_client: Optional[DockerApiClient] = None) -> Optional[Any]:
        """Builds the image using buildx

        Args:
            docker_client: The DockerApiClient for the current cluster

        Options: https://docs.docker.com/engine/reference/commandline/buildx_build/#options
        """
        try:
            import subprocess

            tag = self.get_image_str()
            nocache = self.skip_docker_cache or self.force
            pull = self.pull or self.force

            print_info(f"Building image: {tag}")
            if self.path is not None:
                print_info(f"\t  path: {self.path}")
            if self.dockerfile is not None:
                print_info(f"    dockerfile: {self.dockerfile}")
            if self.platforms is not None:
                print_info(f"     platforms: {self.platforms}")
            logger.debug(f"nocache: {nocache}")
            logger.debug(f"pull: {pull}")

            command = ["docker", "buildx", "build"]

            # Add tag
            command.extend(["--tag", tag])

            # Add dockerfile option, if set
            if self.dockerfile is not None:
                command.extend(["--file", self.dockerfile])

            # Add build arguments
            if self.buildargs:
                for key, value in self.buildargs.items():
                    command.extend(["--build-arg", f"{key}={value}"])

            # Add no-cache option, if set
            if nocache:
                command.append("--no-cache")

            if not self.rm:
                command.append("--rm=false")

            if self.platforms:
                command.append("--platform={}".format(",".join(self.platforms)))

            if self.pull:
                command.append("--pull")

            if self.push_image:
                command.append("--push")
            else:
                command.append("--load")

            # Add path
            if self.path is not None:
                command.append(self.path)

            # Run the command
            logger.debug("Running command: {}".format(" ".join(command)))
            result = subprocess.run(command)

            # Handling output and errors
            if result.returncode == 0:
                print_info("Docker image built successfully.")
                return True
                # _docker_client = docker_client or self.get_docker_client()
                # return self._read(docker_client=_docker_client)
            else:
                logger.error("Error in building Docker image:")
                return False
        except Exception as e:
            logger.error(e)
            return None

    def build_image(self, docker_client: DockerApiClient) -> Optional[Any]:
        if self.platforms is not None or self.use_buildx:
            logger.debug("Using buildx for multi-platform build")
            return self.buildx(docker_client=docker_client)

        from docker import DockerClient
        from docker.errors import APIError, BuildError
        from rich import box
        from rich.live import Live
        from rich.table import Table

        print_info(f"Building image: {self.get_image_str()}")
        nocache = self.skip_docker_cache or self.force
        pull = self.pull or self.force
        if self.path is not None:
            print_info(f"\t  path: {self.path}")
        if self.dockerfile is not None:
            print_info(f"    dockerfile: {self.dockerfile}")
        logger.debug(f"platform: {self.platform}")
        logger.debug(f"nocache: {nocache}")
        logger.debug(f"pull: {pull}")

        last_status = None
        last_build_log = None
        build_log_output: List[Any] = []
        build_step_progress: List[str] = []
        build_log_to_show_on_error: List[str] = []
        try:
            _api_client: DockerClient = docker_client.api_client
            build_stream = _api_client.api.build(
                tag=self.get_image_str(),
                path=self.path,
                dockerfile=self.dockerfile,
                nocache=nocache,
                rm=self.rm,
                forcerm=self.forcerm,
                timeout=self.timeout,
                pull=pull,
                buildargs=self.buildargs,
                container_limits=self.container_limits,
                shmsize=self.shmsize,
                labels=self.labels,
                cache_from=self.cache_from,
                target=self.target,
                network_mode=self.network_mode,
                squash=self.squash,
                extra_hosts=self.extra_hosts,
                platform=self.platform,
                isolation=self.isolation,
                use_config_proxy=self.use_config_proxy,
                decode=True,
            )

            with Live(transient=True, console=console) as live_log:
                for build_log in build_stream:
                    if build_log != last_build_log:
                        last_build_log = build_log
                        build_log_output.append(build_log)

                    build_status: str = build_log.get("status")
                    if build_status is not None:
                        _status = build_status.lower()
                        if _status in (
                            "waiting",
                            "downloading",
                            "extracting",
                            "verifying checksum",
                            "pulling fs layer",
                        ):
                            continue
                        if build_status != last_status:
                            logger.debug(build_status)
                            last_status = build_status

                    if build_log.get("error", None) is not None:
                        live_log.stop()
                        logger.error(build_log_output[-50:])
                        logger.error(build_log["error"])
                        logger.error(f"Image build failed: {self.get_image_str()}")
                        return None

                    stream = build_log.get("stream", None)
                    if stream is None or stream == "\n":
                        continue
                    stream = stream.strip()

                    if "Step" in stream and self.print_build_log:
                        build_step_progress = []
                        print_info(stream)
                    else:
                        build_step_progress.append(stream)
                        if len(build_step_progress) > 10:
                            build_step_progress.pop(0)

                    build_log_to_show_on_error.append(stream)
                    if len(build_log_to_show_on_error) > 50:
                        build_log_to_show_on_error.pop(0)

                    if "error" in stream.lower():
                        print(stream)
                        live_log.stop()

                        # Render error table
                        error_table = Table(show_edge=False, show_header=False, show_lines=False)
                        for line in build_log_to_show_on_error:
                            error_table.add_row(line, style="dim")
                        error_table.add_row(stream, style="bold red")
                        console.print(error_table)
                        return None
                    if build_log.get("aux", None) is not None:
                        logger.debug("build_log['aux'] :{}".format(build_log["aux"]))
                        self.image_build_id = build_log.get("aux", {}).get("ID")

                    # Render table
                    table = Table(show_edge=False, show_header=False, show_lines=False)
                    for line in build_step_progress:
                        table.add_row(line, style="dim")
                    live_log.update(table)

            if self.push_image:
                print_info(f"Pushing {self.get_image_str()}")
                with Live(transient=True, console=console) as live_log:
                    push_status = {}
                    last_push_progress = None
                    for push_output in _api_client.images.push(
                        repository=self.name,
                        tag=self.tag,
                        stream=True,
                        decode=True,
                    ):
                        _id = push_output.get("id", None)
                        _status = push_output.get("status", None)
                        _progress = push_output.get("progress", None)
                        if _id is not None and _status is not None:
                            push_status[_id] = {
                                "status": _status,
                                "progress": _progress,
                            }

                        if push_output.get("error", None) is not None:
                            logger.error(push_output["error"])
                            logger.error(f"Push failed for {self.get_image_str()}")
                            logger.error("If you are using a private registry, make sure you are logged in")
                            return None

                        if self.print_push_output and push_output.get("status", None) in (
                            "Pushing",
                            "Pushed",
                        ):
                            current_progress = push_output.get("progress", None)
                            if current_progress != last_push_progress:
                                print_info(current_progress)
                                last_push_progress = current_progress
                        if push_output.get("aux", {}).get("Size", 0) > 0:
                            print_info(f"Push complete: {push_output.get('aux', {})}")

                        # Render table
                        table = Table(box=box.ASCII2)
                        table.add_column("Layer", justify="center")
                        table.add_column("Status", justify="center")
                        table.add_column("Progress", justify="center")
                        for layer, layer_status in push_status.items():
                            table.add_row(
                                layer,
                                layer_status["status"],
                                layer_status["progress"],
                                style="dim",
                            )
                        live_log.update(table)

            return self._read(docker_client)
        except TypeError as type_error:
            logger.error(type_error)
        except BuildError as build_error:
            logger.error(build_error)
        except APIError as api_err:
            logger.error(api_err)
        except Exception as e:
            logger.error(e)
        return None

    def _create(self, docker_client: DockerApiClient) -> bool:
        """Creates the image

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        logger.debug("Creating: {}".format(self.get_resource_name()))
        try:
            image_object = self.build_image(docker_client)
            if image_object is not None:
                return True
            return False
            # if image_object is not None and isinstance(image_object, Image):
            #     logger.debug("Image built: {}".format(image_object))
            #     self.active_resource = image_object
            #     return True
        except Exception as e:
            logger.exception(e)
            logger.error("Error while creating image: {}".format(e))
            raise

    def _read(self, docker_client: DockerApiClient) -> Any:
        """Returns an Image object if available

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        from docker import DockerClient
        from docker.errors import ImageNotFound, NotFound
        from docker.models.images import Image

        logger.debug("Reading: {}".format(self.get_image_str()))
        try:
            _api_client: DockerClient = docker_client.api_client
            image_object: Optional[List[Image]] = _api_client.images.get(name=self.get_image_str())
            if image_object is not None and isinstance(image_object, Image):
                logger.debug("Image found: {}".format(image_object))
                self.active_resource = image_object
                return image_object
        except (NotFound, ImageNotFound):
            logger.debug(f"Image {self.tag} not found")

        return None

    def _update(self, docker_client: DockerApiClient) -> bool:
        """Updates the Image

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        logger.debug("Updating: {}".format(self.get_resource_name()))
        return self._create(docker_client=docker_client)

    def _delete(self, docker_client: DockerApiClient) -> bool:
        """Deletes the Image

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        from docker import DockerClient
        from docker.models.images import Image

        logger.debug("Deleting: {}".format(self.get_resource_name()))
        image_object: Optional[Image] = self._read(docker_client)
        # Return True if there is no image to delete
        if image_object is None:
            logger.debug("No image to delete")
            return True

        # Delete Image
        try:
            self.active_resource = None
            logger.debug("Deleting image: {}".format(self.tag))
            _api_client: DockerClient = docker_client.api_client
            _api_client.images.remove(image=self.tag, force=True)
            return True
        except Exception as e:
            logger.exception("Error while deleting image: {}".format(e))

        return False

    def create(self, docker_client: DockerApiClient) -> bool:
        # If self.force then always create container
        if not self.force:
            # If use_cache is True and image is active then return True
            if self.use_cache and self.is_active(docker_client=docker_client):
                print_info(f"{self.get_resource_type()}: {self.get_resource_name()} already exists")
                return True

        resource_created = self._create(docker_client=docker_client)
        if resource_created:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} created")
            return True
        logger.error(f"Failed to create {self.get_resource_type()}: {self.get_resource_name()}")
        return False
